""" A sparse matrix in COOrdinate or 'triplet' format"""

__docformat__ = "restructuredtext en"

__all__ = ['coo_array', 'coo_matrix', 'isspmatrix_coo']

import math
from warnings import warn

import numpy as np

from .._lib._util import copy_if_needed
from ._matrix import spmatrix
from ._sparsetools import (coo_tocsr, coo_todense, coo_todense_nd,
                           coo_matvec, coo_matvec_nd, coo_matmat_dense,
                           coo_matmat_dense_nd)
from ._base import issparse, SparseEfficiencyWarning, _spbase, sparray
from ._data import _data_matrix, _minmax_mixin
from ._sputils import (upcast_char, to_native, isshape, getdtype,
                       getdata, downcast_intp_index, get_index_dtype,
                       check_shape, isscalarlike,
                       isintlike, isdense)
from ._index import _validate_indices, _broadcast_arrays

import operator


class _coo_base(_data_matrix, _minmax_mixin):
    _format = 'coo'
    _allow_nd = tuple(range(1, 65))

    def __init__(self, arg1, shape=None, dtype=None, copy=False, *, maxprint=None):
        _data_matrix.__init__(self, arg1, maxprint=maxprint)
        if not copy:
            copy = copy_if_needed

        if isinstance(arg1, tuple):
            if isshape(arg1, allow_nd=self._allow_nd):
                self._shape = check_shape(arg1, allow_nd=self._allow_nd)
                idx_dtype = self._get_index_dtype(maxval=max(self._shape))
                data_dtype = getdtype(dtype, default=float)
                self.coords = tuple(np.array([], dtype=idx_dtype)
                                     for _ in range(len(self._shape)))
                self.data = np.array([], dtype=data_dtype)
                self.has_canonical_format = True
            else:
                try:
                    obj, coords = arg1
                except (TypeError, ValueError) as e:
                    raise TypeError('invalid input format') from e

                if shape is None:
                    if any(len(idx) == 0 for idx in coords):
                        raise ValueError('cannot infer dimensions from zero '
                                         'sized index arrays')
                    shape = tuple(operator.index(np.max(idx)) + 1
                                  for idx in coords)
                self._shape = check_shape(shape, allow_nd=self._allow_nd)
                idx_dtype = self._get_index_dtype(coords,
                                                  maxval=max(self.shape),
                                                  check_contents=True)
                self.coords = tuple(np.array(idx, copy=copy, dtype=idx_dtype)
                                     for idx in coords)
                self.data = getdata(obj, copy=copy, dtype=dtype)
                self.has_canonical_format = False
        else:
            if issparse(arg1):
                if arg1.format == self.format and copy:
                    self.coords = tuple(idx.copy() for idx in arg1.coords)
                    self.data = arg1.data.astype(getdtype(dtype, arg1))  # copy=True
                    self._shape = check_shape(arg1.shape, allow_nd=self._allow_nd)
                    self.has_canonical_format = arg1.has_canonical_format
                else:
                    coo = arg1.tocoo(copy=copy)
                    self.coords = tuple(coo.coords)
                    self.data = coo.data.astype(getdtype(dtype, coo), copy=False)
                    self._shape = check_shape(coo.shape, allow_nd=self._allow_nd)
                    self.has_canonical_format = False
            else:
                # dense argument
                M = np.asarray(arg1)
                if not isinstance(self, sparray):
                    M = np.atleast_2d(M)
                    if M.ndim != 2:
                        raise TypeError(f'expected 2D array or matrix, not {M.ndim}D')

                self._shape = check_shape(M.shape, allow_nd=self._allow_nd)
                if shape is not None:
                    if check_shape(shape, allow_nd=self._allow_nd) != self._shape:
                        message = f'inconsistent shapes: {shape} != {self._shape}'
                        raise ValueError(message)

                index_dtype = self._get_index_dtype(maxval=max(self._shape))
                coords = M.nonzero()
                self.coords = tuple(idx.astype(index_dtype, copy=False)
                                     for idx in coords)
                self.data = getdata(M[coords], copy=copy, dtype=dtype)
                self.has_canonical_format = True

        if len(self._shape) > 2:
            self.coords = tuple(idx.astype(np.int64, copy=False) for idx in self.coords)

        self._check()

    @property
    def row(self):
        if self.ndim > 1:
            return self.coords[-2]
        result = np.zeros_like(self.col)
        result.setflags(write=False)
        return result


    @row.setter
    def row(self, new_row):
        if self.ndim < 2:
            raise ValueError('cannot set row attribute of a 1-dimensional sparse array')
        new_row = np.asarray(new_row, dtype=self.coords[-2].dtype)
        self.coords = self.coords[:-2] + (new_row,) + self.coords[-1:]

    @property
    def col(self):
        return self.coords[-1]

    @col.setter
    def col(self, new_col):
        new_col = np.asarray(new_col, dtype=self.coords[-1].dtype)
        self.coords = self.coords[:-1] + (new_col,)

    def reshape(self, *shape, order="C", copy=False):
        shape = check_shape(shape, self.shape, allow_nd=self._allow_nd)

        # Return early if reshape is not required
        if shape == self.shape:
            if copy:
                return self.copy()
            else:
                return self

        # When reducing the number of dimensions, we need to be careful about
        # index overflow. This is why we can't simply call
        # `np.ravel_multi_index()` followed by `np.unravel_index()` here.
        flat_coords = _ravel_coords(self.coords, self.shape, order=order)
        if len(shape) == 2:
            if order == 'C':
                new_coords = divmod(flat_coords, shape[1])
            else:
                new_coords = divmod(flat_coords, shape[0])[::-1]
        else:
            new_coords = np.unravel_index(flat_coords, shape, order=order)

        idx_dtype = self._get_index_dtype(self.coords, maxval=max(shape))
        new_coords = tuple(np.asarray(co, dtype=idx_dtype) for co in new_coords)

        # Handle copy here rather than passing on to the constructor so that no
        # copy will be made of `new_coords` regardless.
        if copy:
            new_data = self.data.copy()
        else:
            new_data = self.data

        return self.__class__((new_data, new_coords), shape=shape, copy=False)

    reshape.__doc__ = _spbase.reshape.__doc__

    def _getnnz(self, axis=None):
        if axis is None or (axis == 0 and self.ndim == 1):
            nnz = len(self.data)
            if any(len(idx) != nnz for idx in self.coords):
                raise ValueError('all index and data arrays must have the '
                                 'same length')

            if self.data.ndim != 1 or any(idx.ndim != 1 for idx in self.coords):
                raise ValueError('coordinates and data arrays must be 1-D')

            return int(nnz)

        if axis < 0:
            axis += self.ndim
        if axis >= self.ndim:
            raise ValueError('axis out of bounds')

        return np.bincount(downcast_intp_index(self.coords[1 - axis]),
                           minlength=self.shape[1 - axis])

    _getnnz.__doc__ = _spbase._getnnz.__doc__

    def count_nonzero(self, axis=None):
        self.sum_duplicates()
        if axis is None:
            return np.count_nonzero(self.data)

        if axis < 0:
            axis += self.ndim
        if axis < 0 or axis >= self.ndim:
            raise ValueError('axis out of bounds')
        mask = self.data != 0
        coord = self.coords[1 - axis][mask]
        return np.bincount(downcast_intp_index(coord), minlength=self.shape[1 - axis])

    count_nonzero.__doc__ = _spbase.count_nonzero.__doc__

    def _check(self):
        """ Checks data structure for consistency """
        if self.ndim != len(self.coords):
            raise ValueError('mismatching number of index arrays for shape; '
                             f'got {len(self.coords)}, expected {self.ndim}')

        # index arrays should have integer data types
        for i, idx in enumerate(self.coords):
            if idx.dtype.kind != 'i':
                warn(f'index array {i} has non-integer dtype ({idx.dtype.name})',
                     stacklevel=3)

        idx_dtype = self._get_index_dtype(self.coords, maxval=max(self.shape))
        self.coords = tuple(np.asarray(idx, dtype=idx_dtype)
                             for idx in self.coords)
        self.data = to_native(self.data)

        if self.nnz > 0:
            for i, idx in enumerate(self.coords):
                if idx.max() >= self.shape[i]:
                    raise ValueError(f'axis {i} index {idx.max()} exceeds '
                                     f'matrix dimension {self.shape[i]}')
                if idx.min() < 0:
                    raise ValueError(f'negative axis {i} index: {idx.min()}')

    def transpose(self, axes=None, copy=False):
        if axes is None:
            axes = range(self.ndim)[::-1]
        elif isinstance(self, sparray):
            if not hasattr(axes, "__len__") or len(axes) != self.ndim:
                raise ValueError("axes don't match matrix dimensions")
            if len(set(axes)) != self.ndim:
                raise ValueError("repeated axis in transpose")
        elif axes != (1, 0):
            raise ValueError("Sparse matrices do not support an 'axes' "
                             "parameter because swapping dimensions is the "
                             "only logical permutation.")

        permuted_shape = tuple(self._shape[i] for i in axes)
        permuted_coords = tuple(self.coords[i] for i in axes)
        return self.__class__((self.data, permuted_coords),
                              shape=permuted_shape, copy=copy)

    transpose.__doc__ = _spbase.transpose.__doc__
    
    @property
    def mT(self):
        if (n := self.ndim) < 2:
            raise ValueError(f"Array must be at least 2-dimensional, but it is {n}-D")
        axes = None if n == 2 else tuple(range(n - 2)) + (-1, -2)
        return self.transpose(axes=axes)
    
    mT.__doc__ = _spbase.mT.__doc__

    def resize(self, *shape) -> None:
        shape = check_shape(shape, allow_nd=self._allow_nd)
        if self.ndim > 2:
            raise ValueError("only 1-D or 2-D input accepted")
        if len(shape) > 2:
            raise ValueError("shape argument must be 1-D or 2-D")
        # Check for added dimensions.
        if len(shape) > self.ndim:
            flat_coords = _ravel_coords(self.coords, self.shape)
            max_size = math.prod(shape)
            self.coords = np.unravel_index(flat_coords[:max_size], shape)
            self.data = self.data[:max_size]
            self._shape = shape
            return

        # Check for removed dimensions.
        if len(shape) < self.ndim:
            tmp_shape = (
                self._shape[:len(shape) - 1]  # Original shape without last axis
                + (-1,)  # Last axis is used to flatten the array
                + (1,) * (self.ndim - len(shape))  # Pad with ones
            )
            tmp = self.reshape(tmp_shape)
            self.coords = tmp.coords[:len(shape)]
            self._shape = tmp.shape[:len(shape)]

        # Handle truncation of existing dimensions.
        is_truncating = any(old > new for old, new in zip(self.shape, shape))
        if is_truncating:
            mask = np.logical_and.reduce([
                idx < size for idx, size in zip(self.coords, shape)
            ])
            if not mask.all():
                self.coords = tuple(idx[mask] for idx in self.coords)
                self.data = self.data[mask]

        self._shape = shape

    resize.__doc__ = _spbase.resize.__doc__

    def toarray(self, order=None, out=None):
        B = self._process_toarray_args(order, out)
        fortran = int(B.flags.f_contiguous)
        if not fortran and not B.flags.c_contiguous:
            raise ValueError("Output array must be C or F contiguous")
        # This handles both 0D and 1D cases correctly regardless of the
        # original shape.
        if self.ndim == 1:
            coo_todense_nd(np.array([1]), self.nnz, self.ndim,
                           self.coords[0], self.data, B.ravel('A'), fortran)
        elif self.ndim == 2:
            M, N = self.shape
            coo_todense(M, N, self.nnz, self.row, self.col, self.data,
                        B.ravel('A'), fortran)
        else:  # dim>2
            if fortran:
                strides = np.append(1, np.cumprod(self.shape[:-1]))
            else:
                strides = np.append(np.cumprod(self.shape[1:][::-1])[::-1], 1)
            coords = np.concatenate(self.coords)
            coo_todense_nd(strides, self.nnz, self.ndim,
                           coords, self.data, B.ravel('A'), fortran)
        # Note: reshape() doesn't copy here, but does return a new array (view).
        return B.reshape(self.shape)

    toarray.__doc__ = _spbase.toarray.__doc__

    def tocsc(self, copy=False):
        """Convert this array/matrix to Compressed Sparse Column format.

        Duplicate entries will be summed together.

        Parameters
        ----------
        copy : bool, optional
            Unused. A copy is always made in the 2D case. And CSC is 2D.

        Returns
        -------
        csc array/matrix
            The converted array/matrix in CSC format.

        Examples
        --------
        >>> from numpy import array
        >>> from scipy.sparse import coo_array
        >>> row  = array([0, 0, 1, 3, 1, 0, 0])
        >>> col  = array([0, 2, 1, 3, 1, 0, 0])
        >>> data = array([1, 1, 1, 1, 1, 1, 1])
        >>> A = coo_array((data, (row, col)), shape=(4, 4)).tocsc()
        >>> A.toarray()
        array([[3, 0, 1, 0],
               [0, 2, 0, 0],
               [0, 0, 0, 0],
               [0, 0, 0, 1]])

        """
        if self.ndim != 2:
            raise ValueError(f'Cannot convert. CSC format must be 2D. Got {self.ndim}D')
        if self.nnz == 0:
            return self._csc_container(self.shape, dtype=self.dtype)
        else:
            # _coo_to_compressed copy kwarg only for 1D. CSC cant be 1D. See gh-24676
            from ._csc import csc_array
            indptr, indices, data, shape = self._coo_to_compressed(csc_array._swap)

            x = self._csc_container((data, indices, indptr), shape=shape)
            if not self.has_canonical_format:
                x.sum_duplicates()
            return x

    def tocsr(self, copy=False):
        """Convert this array/matrix to Compressed Sparse Row format.

        Duplicate entries will be summed together.

        Parameters
        ----------
        copy : bool, optional
            With ``copy=False``, the data/indices may be shared between this
            array/matrix and the resultant csr_array/matrix. But only for 1D.
            For 2D, a copy will always be made.

        Returns
        -------
        csr array/matrix
            The converted array/matrix in CSR format.

        Examples
        --------
        >>> from numpy import array
        >>> from scipy.sparse import coo_array
        >>> row  = array([0, 0, 1, 3, 1, 0, 0])
        >>> col  = array([0, 2, 1, 3, 1, 0, 0])
        >>> data = array([1, 1, 1, 1, 1, 1, 1])
        >>> A = coo_array((data, (row, col)), shape=(4, 4)).tocsr()
        >>> A.toarray()
        array([[3, 0, 1, 0],
               [0, 2, 0, 0],
               [0, 0, 0, 0],
               [0, 0, 0, 1]])

        """
        if self.ndim > 2:
            raise ValueError(f'Cannot convert. CSR must be 1D or 2D. Got {self.ndim}D')
        if self.nnz == 0:
            return self._csr_container(self.shape, dtype=self.dtype)
        else:
            from ._csr import csr_array
            arrays = self._coo_to_compressed(csr_array._swap, copy=copy)
            indptr, indices, data, shape = arrays

            x = self._csr_container((data, indices, indptr), shape=self.shape)
            if not self.has_canonical_format:
                x.sum_duplicates()
            return x

    def _coo_to_compressed(self, swap, copy=False):
        """convert (shape, coords, data) to (indptr, indices, data, shape)"""
        M, N = swap(self._shape_as_2d)
        # convert idx_dtype intc to int32 for pythran.
        # tested in scipy/optimize/tests/test__numdiff.py::test_group_columns
        idx_dtype = self._get_index_dtype(self.coords, maxval=max(self.nnz, N))

        if self.ndim == 1:
            indices = self.coords[0].copy() if copy else self.coords[0]
            nnz = len(indices)
            indptr = np.array([0, nnz], dtype=idx_dtype)
            data = self.data.copy() if copy else self.data
            return indptr, indices, data, self.shape

        # ndim == 2
        major, minor = swap(self.coords)
        nnz = len(major)
        major = major.astype(idx_dtype, copy=False)
        minor = minor.astype(idx_dtype, copy=False)

        indptr = np.empty(M + 1, dtype=idx_dtype)
        indices = np.empty_like(minor, dtype=idx_dtype)
        data = np.empty_like(self.data, dtype=self.dtype)

        coo_tocsr(M, N, nnz, major, minor, self.data, indptr, indices, data)
        return indptr, indices, data, self.shape

    def tocoo(self, copy=False):
        if copy:
            return self.copy()
        else:
            return self

    tocoo.__doc__ = _spbase.tocoo.__doc__

    def todia(self, copy=False):
        if self.ndim != 2:
            raise ValueError(f'Cannot convert. DIA format must be 2D. Got {self.ndim}D')
        self.sum_duplicates()
        ks = self.col - self.row  # the diagonal for each nonzero
        diags, diag_idx = np.unique(ks, return_inverse=True)

        if len(diags) > 100:
            # probably undesired, should todia() have a maxdiags parameter?
            warn(f"Constructing a DIA matrix with {len(diags)} diagonals "
                 "is inefficient",
                 SparseEfficiencyWarning, stacklevel=2)

        #initialize and fill in data array
        if self.data.size == 0:
            data = np.zeros((0, 0), dtype=self.dtype)
        else:
            data = np.zeros((len(diags), self.col.max()+1), dtype=self.dtype)
            data[diag_idx, self.col] = self.data

        return self._dia_container((data, diags), shape=self.shape)

    todia.__doc__ = _spbase.todia.__doc__

    def todok(self, copy=False):
        if self.ndim > 2:
            raise ValueError(f'Cannot convert. DOK must be 1D or 2D. Got {self.ndim}D')
        self.sum_duplicates()
        dok = self._dok_container(self.shape, dtype=self.dtype)
        # ensure that 1d coordinates are not tuples
        if self.ndim == 1:
            coords = self.coords[0]
        else:
            coords = zip(*self.coords)

        dok._dict = dict(zip(coords, self.data))
        return dok

    todok.__doc__ = _spbase.todok.__doc__

    def diagonal(self, k=0):
        if self.ndim != 2:
            raise ValueError("diagonal requires two dimensions")
        rows, cols = self.shape
        if k <= -rows or k >= cols:
            return np.empty(0, dtype=self.data.dtype)
        diag = np.zeros(min(rows + min(k, 0), cols - max(k, 0)),
                        dtype=self.dtype)
        diag_mask = (self.row + k) == self.col

        if self.has_canonical_format:
            row = self.row[diag_mask]
            data = self.data[diag_mask]
        else:
            inds = tuple(idx[diag_mask] for idx in self.coords)
            (row, _), data = self._sum_duplicates(inds, self.data[diag_mask])
        diag[row + min(k, 0)] = data

        return diag

    diagonal.__doc__ = _data_matrix.diagonal.__doc__

    def _setdiag(self, values, k):
        if self.ndim != 2:
            raise ValueError("setting a diagonal requires two dimensions")
        M, N = self.shape
        if values.ndim and not len(values):
            return
        idx_dtype = self.row.dtype

        # Determine which triples to keep and where to put the new ones.
        full_keep = self.col - self.row != k
        if k < 0:
            max_index = min(M+k, N)
            if values.ndim:
                max_index = min(max_index, len(values))
            keep = np.logical_or(full_keep, self.col >= max_index)
            new_row = np.arange(-k, -k + max_index, dtype=idx_dtype)
            new_col = np.arange(max_index, dtype=idx_dtype)
        else:
            max_index = min(M, N-k)
            if values.ndim:
                max_index = min(max_index, len(values))
            keep = np.logical_or(full_keep, self.row >= max_index)
            new_row = np.arange(max_index, dtype=idx_dtype)
            new_col = np.arange(k, k + max_index, dtype=idx_dtype)

        # Define the array of data consisting of the entries to be added.
        if values.ndim:
            new_data = values[:max_index]
        else:
            new_data = np.empty(max_index, dtype=self.dtype)
            new_data[:] = values

        # Update the internal structure.
        self.coords = (np.concatenate((self.row[keep], new_row)),
                       np.concatenate((self.col[keep], new_col)))
        self.data = np.concatenate((self.data[keep], new_data))
        self.has_canonical_format = False

    # needed by _data_matrix
    def _with_data(self, data, copy=True):
        """Returns a matrix with the same sparsity structure as self,
        but with different data. By default the index arrays are copied.
        """
        if copy:
            coords = tuple(idx.copy() for idx in self.coords)
        else:
            coords = self.coords
        return self.__class__((data, coords), shape=self.shape, dtype=data.dtype)

    def __getitem__(self, key):
        index, new_shape, arr_int_pos, none_pos = _validate_indices(
            key, self.shape, self.format
        )
        # handle int, slice and int-array indices
        index_mask = np.ones(len(self.data), dtype=np.bool_)
        slice_coords = []
        arr_coords = []
        arr_indices = []
        for i, (idx, co) in enumerate(zip(index, self.coords)):
            if isinstance(idx, int):
                index_mask &= (co == idx)
            elif isinstance(idx, slice):
                if idx == slice(None):
                    slice_coords.append(co)
                else:
                    start, stop, step = idx.indices(self.shape[i])
                    if step != 1:
                        if step < 0:
                            in_range = (co <= start) & (co > stop)
                        else:
                            in_range = (co >= start) & (co < stop)
                        new_ix, m = np.divmod(co - start, step)
                        index_mask &= (m == 0) & in_range
                    else:
                        in_range = (co >= start) & (co < stop)
                        new_ix = co - start
                        index_mask &= in_range
                    slice_coords.append(new_ix)
            else:  # array
                arr_coords.append(co)
                arr_indices.append(idx)
        # shortcut for scalar output
        if new_shape == ():
            return self.data[index_mask].sum().astype(self.dtype, copy=False)

        new_coords = [co[index_mask] for co in slice_coords]
        new_data = self.data[index_mask]

        # handle array indices
        if arr_indices:
            arr_indices = _broadcast_arrays(*arr_indices)
            arr_shape = arr_indices[0].shape  # already broadcast in validate_indices
            # There are three dimensions required to check array indices against coords
            # Their lengths are described as:
            # a) number of indices that are arrays - arr_dim
            # b) number of coords to check - masked_nnz (already masked by slices)
            # c) size of the index arrays - arr_size
            # Note for this, integer indices are treated like slices, not like arrays.
            #
            # Goal:
            # Find new_coords and index positions that match across all arr_dim axes.
            # Approach: Track matches using bool array. Size: masked_nnz by arr_size.
            # True means all arr_indices match at that coord and index position.
            # Equate with broadcasting and check for all equal across (arr_dim) axis 0.
            # 1st array is "keyarr" (arr_dim by 1 by arr_size) from arr_indices.
            # 2nd array is "arr_coords" (arr_dim by masked_nnz by 1) from arr_coords.
            keyarr = np.array(arr_indices).reshape(len(arr_indices), 1, -1)
            arr_coords = np.array([co[index_mask] for co in arr_coords])[:, :, None]
            found = (keyarr == arr_coords).all(axis=0)
            arr_co, arr_ix = found.nonzero()
            new_data = new_data[arr_co]
            new_coords = [co[arr_co] for co in new_coords]
            new_arr_coords = list(np.unravel_index(arr_ix, shape=arr_shape))

            # check for contiguous positions of array and int indices
            if len(arr_int_pos) == arr_int_pos[-1] - arr_int_pos[0] + 1:
                # Contiguous. Put all array index shape at pos of array indices
                pos = arr_int_pos[0]
                new_coords = new_coords[:pos] + new_arr_coords + new_coords[pos:]
            else:
                # Not contiguous. Put all array coords at front
                new_coords = new_arr_coords + new_coords

        if none_pos:
            if new_coords:
                coord_like = np.zeros_like(new_coords[0])
            else:
                coord_like = np.zeros(len(new_data), dtype=self.coords[0].dtype)
            new_coords.insert(none_pos[0], coord_like)
            for i in none_pos[1:]:
                new_coords.insert(i, coord_like.copy())
        return coo_array((new_data, new_coords), shape=new_shape, dtype=self.dtype)

    def __setitem__(self, key, x):
        # enact self[key] = x
        index, new_shape, arr_int_pos, none_pos = _validate_indices(
            key, self.shape, self.format
        )

        # remove None's at beginning of index. Should not impact indexing coords
        # and will mistakenly align with x_coord columns if not removed.
        if none_pos:
            new_shape = list(new_shape)
            for j in none_pos[::-1]:
                new_shape.pop(j)
            new_shape = tuple(new_shape)

        # broadcast arrays
        if arr_int_pos:
            index = list(index)
            arr_pos = {i: arr for i in arr_int_pos if not isintlike(arr := index[i])}
            arr_indices = _broadcast_arrays(*arr_pos.values())
            for i, arr in zip(arr_pos, arr_indices):
                index[i] = arr

        # get coords and data from x
        if issparse(x):
            if 0 in x.shape:
                return  # Nothing to set.
            x_data, x_coords = _get_sparse_data_and_coords(x, new_shape, self.dtype)
        else:
            x = np.asarray(x, dtype=self.dtype)
            if x.size == 0:
                return  # Nothing to set.
            x_data, x_coords = _get_dense_data_and_coords(x, new_shape)

        # Approach:
        # Set indexed values to zero (drop from `self.coords` and `self.data`)
        # create new coords and data arrays for setting nonzeros
        # concatenate old (undropped) values with new coords and data

        old_data, old_coords = self._zero_many(index)

        if len(x_coords) == 1 and len(x_coords[0]) == 0:
            self.data, self.coords = old_data, old_coords
            # leave self.has_canonical_format unchanged
            return

        # To process array indices, need the x_coords for those axes
        # and need to ravel the array part of x_coords to build new_coords
        # Along the way, Find pos and shape of array-index portion of key.
        # arr_shape is None and pos = -1 when no arrays are used as indices.
        arr_shape = None
        pos = -1
        if arr_int_pos:
            # Get arr_shape if any arrays are in the index.
            # Also ravel the corresponding x_coords.
            for idx in index:
                if not isinstance(idx, slice) and not isintlike(idx):
                    arr_shape = idx.shape

                    # Find x_coord pos of integer and array portion of index.
                    # If contiguous put int and array axes at pos of those indices.
                    # If not contiguous, put all int and array axes at pos=0.
                    if len(arr_int_pos) == (arr_int_pos[-1] - arr_int_pos[0] + 1):
                        pos = arr_int_pos[0]
                    else:
                        pos = 0

                    # compute the raveled coords of the array part of x_coords.
                    # Used to build the new coords from the index arrays.
                    x_arr_coo = x_coords[pos:pos + len(arr_shape)]
                    # could use np.ravel_multi_index but _ravel_coords avoids overflow
                    x_arr_coo_ravel = _ravel_coords(x_arr_coo, arr_shape)
                    break

        # find map from x_coord slice axes to index axes
        x_ax = 0
        x_axes = {}
        for i, idx in enumerate(index):
            if i == pos:
                x_ax += len(arr_shape)
            if isinstance(idx, slice):
                x_axes[i] = x_ax
                x_ax += 1

        # Build new_coords and new_data
        new_coords = [None] * self.ndim
        new_nnz = len(x_data)
        for i, idx in enumerate(index):
            if isintlike(idx):
                new_coords[i] = (np.broadcast_to(idx, (new_nnz,)))
                continue
            elif isinstance(idx, slice):
                start, stop, step = idx.indices(self.shape[i])
                new_coords[i] = (start + x_coords[x_axes[i]] * step)
            else:  # array idx
                new_coords[i] = idx.ravel()[x_arr_coo_ravel]
        # seems like a copy is prudent when setting data in this array
        new_data = x_data.copy()

        # deduplicate entries created by multiple coords matching in the array index
        # NumPy does not specify which value is put into the spot (last one assigned)
        # so only one value should appear if we want to match NumPy.
        # If matching NumPy is not crucial, we could make dups a feature where
        # integer array indices  with repeated indices creates duplicate values.
        # Not doing that here. We are just removing duplicates (keep 1st data found)
        new_coords = np.array(new_coords)
        _, ind = np.unique(new_coords, axis=1, return_index=True)

        # update values with stack of old and new data and coords.
        self.data = np.hstack([old_data, new_data[ind]])
        self.coords = tuple(np.hstack(c) for c in zip(old_coords, new_coords[:, ind]))
        self.has_canonical_format = False

    def _zero_many(self, index):
        # handle int, slice and integer-array indices
        # index_mask accumulates a bool array of nonzeros that match index
        index_mask=np.ones(len(self.data), dtype=np.bool_)
        arr_coords = []
        arr_indices = []
        for i, (idx, co) in enumerate(zip(index, self.coords)):
            if isinstance(idx, int):
                index_mask &= (co == idx)
            elif isinstance(idx, slice) and idx != slice(None):
                start, stop, step = idx.indices(self.shape[i])
                if step != 1:
                    if step < 0:
                        in_range = (co <= start) & (co > stop)
                    else:
                        in_range = (co >= start) & (co < stop)
                    m = np.mod(co - start, step)
                    index_mask &= (m == 0) & in_range
                else:
                    in_range = (co >= start) & (co < stop)
                    index_mask &= in_range
            elif isinstance(idx, slice) and idx == slice(None):
                # slice is full axis so no changes to index_mask
                pass
            else:  # array
                arr_coords.append(co)
                arr_indices.append(idx)

        # match array indices with masked coords. See comments in __getitem__
        if arr_indices:
            keyarr = np.array(arr_indices).reshape(len(arr_indices), 1, -1)
            arr_coords = np.array([co[index_mask] for co in arr_coords])[:, :, None]
            found = (keyarr == arr_coords).all(axis=0)
            arr_coo, _ = found.nonzero()
            arr_index_mask = np.zeros_like(index_mask)
            arr_index_mask[index_mask.nonzero()[0][arr_coo]] = True
            index_mask &= arr_index_mask

        # remove matching coords and data to set them to zero
        pruned_coords = [co[~index_mask] for co in self.coords]
        pruned_data = self.data[~index_mask]
        return pruned_data, pruned_coords

    def sum_duplicates(self) -> None:
        """Eliminate duplicate entries by adding them together.

        This is an *in place* operation
        """
        if self.has_canonical_format:
            return
        summed = self._sum_duplicates(self.coords, self.data)
        self.coords, self.data = summed
        self.has_canonical_format = True

    def _sum_duplicates(self, coords, data):
        # Assumes coords not in canonical format.
        if len(data) == 0:
            return coords, data
        # Sort coords w.r.t. rows, then cols. This corresponds to C-order,
        # which we rely on for argmin/argmax to return the first index in the
        # same way that numpy does (in the case of ties).
        order = np.lexsort(coords[::-1])
        coords = tuple(idx[order] for idx in coords)
        data = data[order]
        unique_mask = np.logical_or.reduce([
            idx[1:] != idx[:-1] for idx in coords
        ])
        unique_mask = np.append(True, unique_mask)
        coords = tuple(idx[unique_mask] for idx in coords)
        unique_inds, = np.nonzero(unique_mask)
        data = np.add.reduceat(data, downcast_intp_index(unique_inds), dtype=self.dtype)
        return coords, data

    def eliminate_zeros(self):
        """Remove zero entries from the array/matrix.

        This is an *in place* operation.
        """
        mask = self.data != 0
        self.data = self.data[mask]
        self.coords = tuple(idx[mask] for idx in self.coords)

    #######################
    # Arithmetic handlers #
    #######################

    def _add_dense(self, other):
        if other.shape != self.shape:
            raise ValueError(f'Incompatible shapes ({self.shape} and {other.shape})')
        dtype = upcast_char(self.dtype.char, other.dtype.char)
        result = np.array(other, dtype=dtype, copy=True)
        fortran = int(result.flags.f_contiguous)
        if self.ndim == 1:
            coo_todense_nd(np.array([1]), self.nnz, self.ndim,
                           self.coords[0], self.data, result.ravel('A'), fortran)
        elif self.ndim == 2:
            M, N = self._shape_as_2d
            coo_todense(M, N, self.nnz, self.row, self.col, self.data,
                        result.ravel('A'), fortran)
        else:
            if fortran:
                strides = np.append(1, np.cumprod(self.shape[:-1]))
            else:
                strides = np.append(np.cumprod(self.shape[1:][::-1])[::-1], 1)
            coords = np.concatenate(self.coords)
            coo_todense_nd(strides, self.nnz, self.ndim,
                           coords, self.data, result.ravel('A'), fortran)
        return self._container(result, copy=False)


    def _add_sparse(self, other):
        if self.ndim < 3:
            return self.tocsr()._add_sparse(other)

        if other.shape != self.shape:
            raise ValueError(f'Incompatible shapes ({self.shape} and {other.shape})')
        other = self.__class__(other)
        new_data = np.concatenate((self.data, other.data))
        new_coords = tuple(np.concatenate((self.coords, other.coords), axis=1))
        A = self.__class__((new_data, new_coords), shape=self.shape)
        return A

    def _sub_sparse(self, other):
        if self.ndim < 3:
            return self.tocsr()._sub_sparse(other)

        if other.shape != self.shape:
            raise ValueError(f'Incompatible shapes ({self.shape} and {other.shape})')
        other = self.__class__(other)
        new_data = np.concatenate((self.data, -other.data))
        new_coords = tuple(np.concatenate((self.coords, other.coords), axis=1))
        A = coo_array((new_data, new_coords), shape=self.shape)
        return A

    def _matmul_vector(self, other):
        if self.ndim > 2:
            result = np.zeros(math.prod(self.shape[:-1]),
                              dtype=upcast_char(self.dtype.char, other.dtype.char))
            shape = np.array(self.shape)
            strides = np.append(np.cumprod(shape[:-1][::-1])[::-1][1:], 1)
            coords = np.concatenate(self.coords)
            coo_matvec_nd(self.nnz, len(self.shape), strides, coords, self.data,
                          other, result)

            result = result.reshape(self.shape[:-1])
            return result

        # self.ndim <= 2
        result_shape = self.shape[0] if self.ndim > 1 else 1
        result = np.zeros(result_shape,
                          dtype=upcast_char(self.dtype.char, other.dtype.char))
        if self.ndim == 2:
            col = self.col
            row = self.row
        elif self.ndim == 1:
            col = self.coords[0]
            row = np.zeros_like(col)
        else:
            raise NotImplementedError(
                f"coo_matvec not implemented for ndim={self.ndim}")

        coo_matvec(self.nnz, row, col, self.data, other, result)
        # Array semantics return a scalar here, not a single-element array.
        if isinstance(self, sparray) and result_shape == 1:
            return result[0]
        return result

    def _rmatmul_dispatch(self, other):
        if isscalarlike(other):
            return self._mul_scalar(other)
        else:
            # Don't use asarray unless we have to
            try:
                o_ndim = other.ndim
            except AttributeError:
                other = np.asarray(other)
                o_ndim = other.ndim
            perm = tuple(range(o_ndim)[:-2]) + tuple(range(o_ndim)[-2:][::-1])
            tr = other.transpose(perm)

            s_ndim = self.ndim
            perm = tuple(range(s_ndim)[:-2]) + tuple(range(s_ndim)[-2:][::-1])
            ret = self.transpose(perm)._matmul_dispatch(tr)
            if ret is NotImplemented:
                return NotImplemented

            if s_ndim == 1 or o_ndim == 1:
                perm = range(ret.ndim)
            else:
                perm = tuple(range(ret.ndim)[:-2]) + tuple(range(ret.ndim)[-2:][::-1])
            return ret.transpose(perm)

    def _matmul_dispatch(self, other):
        if isscalarlike(other):
            return self.multiply(other)

        if not (issparse(other) or isdense(other)):
            # If it's a list or whatever, treat it like an array
            other_a = np.asanyarray(other)

            if other_a.ndim == 0 and other_a.dtype == np.object_:
                # Not interpretable as an array; return NotImplemented so that
                # other's __rmatmul__ can kick in if that's implemented.
                return NotImplemented
            # Allow custom sparse class indicated by attr sparse gh-6520
            try:
                other.shape
            except AttributeError:
                other = other_a

        if self.ndim < 3 and other.ndim < 3:
            return _spbase._matmul_dispatch(self, other)

        N = self.shape[-1]
        err_prefix = "matmul: dimension mismatch with signature"
        if other.__class__ is np.ndarray:
            if other.shape == (N,):
                return self._matmul_vector(other)
            if other.shape == (N, 1):
                result = self._matmul_vector(other.ravel())
                return result.reshape(*self.shape[:-1], 1)
            if other.ndim == 1:
                msg = f"{err_prefix} (n,k={N}),(k={other.shape[0]},)->(n,)"
                raise ValueError(msg)
            if other.shape[-2] == N:
                # check for batch dimensions compatibility
                batch_shape_A = self.shape[:-2]
                batch_shape_B = other.shape[:-2]
                if batch_shape_A != batch_shape_B:
                    try:
                        # This will raise an error if the shapes are not broadcastable
                        np.broadcast_shapes(batch_shape_A, batch_shape_B)
                    except ValueError:
                        raise ValueError("Batch dimensions are not broadcastable")

                return self._matmul_multivector(other)
            else:
                raise ValueError(
                    f"{err_prefix} (n,..,k={N}),(k={other.shape[-2]},..,m)->(n,..,m)"
                )

        if isscalarlike(other):
            # scalar value
            return self._mul_scalar(other)

        if issparse(other):
            self_is_1d = self.ndim == 1
            other_is_1d = other.ndim == 1

            # reshape to 2-D if self or other is 1-D
            if self_is_1d:
                self = self.reshape(self._shape_as_2d) # prepend 1 to shape

            if other_is_1d:
                other = other.reshape((other.shape[0], 1)) # append 1 to shape

            # Check if the inner dimensions match for matrix multiplication
            if N != other.shape[-2]:
                raise ValueError(
                    f"{err_prefix} (n,..,k={N}),(k={other.shape[-2]},..,m)->(n,..,m)"
                )

            # If A or B has more than 2 dimensions, check for
            # batch dimensions compatibility
            if self.ndim > 2 or other.ndim > 2:
                batch_shape_A = self.shape[:-2]
                batch_shape_B = other.shape[:-2]
                if batch_shape_A != batch_shape_B:
                    try:
                        # This will raise an error if the shapes are not broadcastable
                        np.broadcast_shapes(batch_shape_A, batch_shape_B)
                    except ValueError:
                        raise ValueError("Batch dimensions are not broadcastable")

            result = self._matmul_sparse(other)

            # reshape back if a or b were originally 1-D
            if self_is_1d:
                # if self was originally 1-D, reshape result accordingly
                result = result.reshape(tuple(result.shape[:-2]) +
                                        tuple(result.shape[-1:]))
            if other_is_1d:
                result = result.reshape(result.shape[:-1])
            return result

    def _matmul_multivector(self, other):
        result_dtype = upcast_char(self.dtype.char, other.dtype.char)
        if self.ndim >= 3 or other.ndim >= 3:
            # if self has shape (N,), reshape to (1,N)
            if self.ndim == 1:
                result = self.reshape(1, self.shape[0])._matmul_multivector(other)
                return result.reshape(tuple(other.shape[:-2]) + tuple(other.shape[-1:]))

            broadcast_shape = np.broadcast_shapes(self.shape[:-2], other.shape[:-2])
            self_shape = broadcast_shape + self.shape[-2:]
            other_shape = broadcast_shape + other.shape[-2:]

            self = self._broadcast_to(self_shape)
            other = np.broadcast_to(other, other_shape)
            result_shape = broadcast_shape + self.shape[-2:-1] + other.shape[-1:]
            result = np.zeros(result_shape, dtype=result_dtype)
            coo_matmat_dense_nd(self.nnz, len(self.shape), other.shape[-1],
                                np.array(other_shape), np.array(result_shape),
                                np.concatenate(self.coords),
                                self.data, other.ravel('C'), result)
            return result

        if self.ndim == 2:
            result_shape = (self.shape[0], other.shape[1])
            col = self.col
            row = self.row
        elif self.ndim == 1:
            result_shape = (other.shape[1],)
            col = self.coords[0]
            row = np.zeros_like(col)
        result = np.zeros(result_shape, dtype=result_dtype)
        coo_matmat_dense(self.nnz, other.shape[-1], row, col,
                         self.data, other.ravel('C'), result)
        return result.view(type=type(other))

    def dot(self, other):
        """Return the dot product of two arrays.

        Strictly speaking a dot product involves two vectors.
        But in the sense that an array with ndim >= 1 is a collection
        of vectors, the function computes the collection of dot products
        between each vector in the first array with each vector in the
        second array. The axis upon which the sum of products is performed
        is the last axis of the first array and the second to last axis of
        the second array. If the second array is 1-D, the last axis is used.

        Thus, if both arrays are 1-D, the inner product is returned.
        If both are 2-D, we have matrix multiplication. If `other` is 1-D,
        the sum product is taken along the last axis of each array. If
        `other` is N-D for N>=2, the sum product is over the last axis of
        the first array and the second-to-last axis of the second array.

        Parameters
        ----------
        other : array_like or sparse array
            Second array

        Returns
        -------
        output : array (sparse or dense)
            The dot product of this array with `other`.
            It will be dense/sparse if `other` is dense/sparse.

        Examples
        --------

        >>> import numpy as np
        >>> from scipy.sparse import coo_array
        >>> A = coo_array([[1, 2, 0], [0, 0, 3], [4, 0, 5]])
        >>> v = np.array([1, 0, -1])
        >>> A.dot(v)
        array([ 1, -3, -1], dtype=int64)

        For 2-D arrays it is the matrix product:

        >>> A = coo_array([[1, 0], [0, 1]])
        >>> B = coo_array([[4, 1], [2, 2]])
        >>> A.dot(B).toarray()
        array([[4, 1],
               [2, 2]])

        For 3-D arrays the shape extends unused axes by other unused axes.

        >>> A = coo_array(np.arange(3*4*5*6)).reshape((3,4,5,6))
        >>> B = coo_array(np.arange(3*4*5*6)).reshape((5,4,6,3))
        >>> A.dot(B).shape
        (3, 4, 5, 5, 4, 3)
        """
        # handle non-array input:  lists, ints, etc
        if not (issparse(other) or isdense(other) or isscalarlike(other)):
            # If it's a list or whatever, treat it like an array
            o_array = np.asanyarray(other)

            if o_array.ndim == 0 and o_array.dtype == np.object_:
                raise TypeError(f"dot argument not supported type: '{type(other)}'")
            try:
                other.shape
            except AttributeError:
                other = o_array

        # Handle scalar multiplication
        if isscalarlike(other):
            return self * other

        # other.shape[-2:][0] gets last index of 1d, next to last index of >1d
        if self.shape[-1] != other.shape[-2:][0]:
            raise ValueError(f"shapes {self.shape} and {other.shape}"
                             " are not aligned for n-D dot")

        if self.ndim < 3 and other.ndim < 3:
            return self @ other
        if isdense(other):
            return self._dense_dot(other)
        return self._sparse_dot(other.tocoo())

    def _sparse_dot(self, other):
        # already checked: at least one is >2d, neither scalar, both are coo
        # Ravel non-reduced axes coordinates
        self_2d, s_new_shape = _convert_to_2d(self, [self.ndim - 1])
        other_2d, o_new_shape = _convert_to_2d(other, [max(0, other.ndim - 2)])

        prod = self_2d @ other_2d.T  # routes via 2-D CSR
        prod = prod.tocoo()

        # Combine the shapes of the non-contracted axes
        combined_shape = s_new_shape + o_new_shape

        # Unravel the 2D coordinates to get multi-dimensional coordinates
        coords = []
        new_shapes = (s_new_shape, o_new_shape) if s_new_shape else (o_new_shape,)
        for c, s in zip(prod.coords, new_shapes):
            coords.extend(np.unravel_index(c, s))

        # Construct the resulting COO array with coords and shape
        return coo_array((prod.data, coords), shape=combined_shape)

    def _dense_dot(self, other):
        # already checked: self is >0d, other is dense and >0d
        # Ravel non-reduced axes coordinates
        s_ndim = self.ndim
        if s_ndim <= 2:
            s_new_shape = () if s_ndim == 1 else (self.shape[0],)
            self_2d = self
        else:
            self_2d, s_new_shape = _convert_to_2d(self, [self.ndim - 1])

        o_ndim = other.ndim
        if o_ndim <= 2:
            o_new_shape = () if o_ndim == 1 else (other.shape[-1],)
            other_2d = other
        else:
            o_new_shape = other.shape[:-2] + other.shape[-1:]
            reorder_dims = (o_ndim - 2, *range(o_ndim - 2), o_ndim - 1)
            o_reorg = np.transpose(other, reorder_dims)
            other_2d = o_reorg.reshape((other.shape[-2], math.prod(o_new_shape)))

        prod = self_2d @ other_2d  # routes via 2-D CSR

        # Combine the shapes of the non-contracted axes
        combined_shape = s_new_shape + o_new_shape
        return prod.reshape(combined_shape)


    def tensordot(self, other, axes=2):
        """Return the tensordot product with another array along the given axes.

        The tensordot differs from dot and matmul in that any axis can be
        chosen for each of the first and second array and the sum of the
        products is computed just like for matrix multiplication, only not
        just for the rows of the first times the columns of the second. It
        takes the dot product of the collection of vectors along the specified
        axes.  Here we can even take the sum of the products along two or even
        more axes if desired. So, tensordot is a dot product computation
        applied to arrays of any dimension >= 1. It is like matmul but over
        arbitrary axes for each matrix.

        Given two tensors, `a` and `b`, and the desired axes specified as a
        2-tuple/list/array containing two sequences of axis numbers,
        ``(a_axes, b_axes)``, sum the products of `a`'s and `b`'s elements
        (components) over the axes specified by ``a_axes`` and ``b_axes``.
        The `axes` input can be a single non-negative integer, ``N``;
        if it is, then the last ``N`` dimensions of `a` and the first
        ``N`` dimensions of `b` are summed over.

        Parameters
        ----------
        other : array_like
            Tensor to "dot".
        axes : int or (2,) array_like
            * integer_like
              If an int N, sum over the last N axes of `a` and the first N axes
              of `b` in order. The sizes of the corresponding axes must match.
            * (2,) array_like
              A 2-tuple of sequences of axes to be summed over, the first applying
              to `a`, the second to `b`. The sequences must be the same length.
              The shape of the corresponding axes must match between `a` and `b`.

        Returns
        -------
        output : coo_array
            The tensor dot product of this array with `other`.
            It will be dense/sparse if `other` is dense/sparse.

        See Also
        --------
        dot

        Examples
        --------
        >>> import numpy as np
        >>> import scipy.sparse
        >>> A = scipy.sparse.coo_array([[[2, 3], [0, 0]], [[0, 1], [0, 5]]])
        >>> A.shape
        (2, 2, 2)

        Integer axes N are shorthand for (range(-N, 0), range(0, N)):

        >>> A.tensordot(A, axes=1).toarray()
        array([[[[ 4,  9],
                 [ 0, 15]],
        <BLANKLINE>
                [[ 0,  0],
                 [ 0,  0]]],
        <BLANKLINE>
        <BLANKLINE>
               [[[ 0,  1],
                 [ 0,  5]],
        <BLANKLINE>
                [[ 0,  5],
                 [ 0, 25]]]])
        >>> A.tensordot(A, axes=2).toarray()
        array([[ 4,  6],
               [ 0, 25]])
        >>> A.tensordot(A, axes=3)
        array(39)

        Using tuple for axes:

        >>> a = scipy.sparse.coo_array(np.arange(60).reshape(3,4,5))
        >>> b = np.arange(24).reshape(4,3,2)
        >>> c = a.tensordot(b, axes=([1,0],[0,1]))
        >>> c.shape
        (5, 2)
        >>> c
        array([[4400, 4730],
               [4532, 4874],
               [4664, 5018],
               [4796, 5162],
               [4928, 5306]])

        """
        if not isdense(other) and not issparse(other):
            # If it's a list or whatever, treat it like an array
            other_array = np.asanyarray(other)

            if other_array.ndim == 0 and other_array.dtype == np.object_:
                raise TypeError(f"tensordot arg not supported type: '{type(other)}'")
            try:
                other.shape
            except AttributeError:
                other = other_array

        axes_self, axes_other = _process_axes(self.ndim, other.ndim, axes)

        # Check for shape compatibility along specified axes
        if any(self.shape[ax] != other.shape[bx]
               for ax, bx in zip(axes_self, axes_other)):
            raise ValueError("sizes of the corresponding axes must match")

        if isdense(other):
            return self._dense_tensordot(other, axes_self, axes_other)
        else:
            return self._sparse_tensordot(other, axes_self, axes_other)

    def _sparse_tensordot(self, other, s_axes, o_axes):
        # Prepare the tensors for tensordot operation
        # Ravel non-reduced axes coordinates
        self_2d, s_new_shape = _convert_to_2d(self, s_axes)
        other_2d, o_new_shape = _convert_to_2d(other, o_axes)

        # Perform matrix multiplication (routed via 2-D CSR)
        prod = self_2d @ other_2d.T
        # handle case of scalar result (axis includes all axes for both)
        if not issparse(prod):
            return prod
        prod = prod.tocoo()

        # Combine the shapes of the non-contracted axes
        combined_shape = s_new_shape + o_new_shape

        # Unravel the 2D coordinates to get multi-dimensional coordinates
        coords = []
        new_shapes = (s_new_shape, o_new_shape) if s_new_shape else (o_new_shape,)
        for c, s in zip(prod.coords, new_shapes):
            if s:
                coords.extend(np.unravel_index(c, s))

        # Construct the resulting COO array with coords and shape
        return coo_array((prod.data, coords), shape=combined_shape)

    def _dense_tensordot(self, other, s_axes, o_axes):
        s_ndim = len(self.shape)
        o_ndim = len(other.shape)

        s_non_axes = [i for i in range(s_ndim) if i not in s_axes]
        s_axes_shape = [self.shape[i] for i in s_axes]
        s_non_axes_shape = [self.shape[i] for i in s_non_axes]

        o_non_axes = [i for i in range(o_ndim) if i not in o_axes]
        o_axes_shape = [other.shape[i] for i in o_axes]
        o_non_axes_shape = [other.shape[i] for i in o_non_axes]

        left = self.transpose(s_non_axes + s_axes)
        right = np.transpose(other, o_non_axes[:-1] + o_axes + o_non_axes[-1:])

        reshape_left = (*s_non_axes_shape, math.prod(s_axes_shape))
        reshape_right = (*o_non_axes_shape[:-1], math.prod(o_axes_shape),
                         *o_non_axes_shape[-1:])

        return left.reshape(reshape_left).dot(right.reshape(reshape_right))

    def _matmul_sparse(self, other):
        """
        Perform sparse-sparse matrix multiplication for two n-D COO arrays.
        The method converts input n-D arrays to 2-D block array format,
        uses csr_matmat to multiply them, and then converts the
        result back to n-D COO array.

        Parameters:
        self (COO): The first n-D sparse array in COO format.
        other (COO): The second n-D sparse array in COO format.

        Returns:
        prod (COO): The resulting n-D sparse array after multiplication.
        """
        if self.ndim < 3 and other.ndim < 3:
            return _spbase._matmul_sparse(self, other)

        # Get the shapes of self and other
        self_shape = self.shape
        other_shape = other.shape

        # Determine the new shape to broadcast self and other
        broadcast_shape = np.broadcast_shapes(self_shape[:-2], other_shape[:-2])
        self_new_shape = tuple(broadcast_shape) + self_shape[-2:]
        other_new_shape = tuple(broadcast_shape) + other_shape[-2:]

        self_broadcasted = self._broadcast_to(self_new_shape)
        other_broadcasted = other._broadcast_to(other_new_shape)

        # Convert n-D COO arrays to 2-D block diagonal arrays
        self_block_diag = _block_diag(self_broadcasted)
        other_block_diag = _block_diag(other_broadcasted)

        # Use csr_matmat to perform sparse matrix multiplication
        prod_block_diag = (self_block_diag @ other_block_diag).tocoo()

        # Convert the 2-D block diagonal array back to n-D
        return _extract_block_diag(
            prod_block_diag,
            shape=(*broadcast_shape, self.shape[-2], other.shape[-1]),
        )

    def _broadcast_to(self, new_shape, copy=False):
        if self.shape == new_shape:
            return self.copy() if copy else self

        old_shape = self.shape

        # Check if the new shape is compatible for broadcasting
        if len(new_shape) < len(old_shape):
            raise ValueError("New shape must have at least as many dimensions"
                             " as the current shape")

        # Add leading ones to shape to ensure same length as `new_shape`
        shape = (1,) * (len(new_shape) - len(old_shape)) + tuple(old_shape)

        # Ensure the old shape can be broadcast to the new shape
        if any((o != 1 and o != n) for o, n in zip(shape, new_shape)):
            raise ValueError(f"current shape {old_shape} cannot be "
                             "broadcast to new shape {new_shape}")

        # Reshape the COO array to match the new dimensions
        self = self.reshape(shape)

        idx_dtype = get_index_dtype(self.coords, maxval=max(new_shape))
        coords = self.coords
        new_data = self.data
        new_coords = coords[-1:]  # Copy last coordinate to start
        cum_repeat = 1 # Cumulative repeat factor for broadcasting

        if shape[-1] != new_shape[-1]: # broadcasting the n-th (col) dimension
            repeat_count = new_shape[-1]
            cum_repeat *= repeat_count
            new_data = np.tile(new_data, repeat_count)
            new_dim = np.repeat(np.arange(0, repeat_count, dtype=idx_dtype), self.nnz)
            new_coords = (new_dim,)

        for i in range(-2, -(len(shape)+1), -1):
            if shape[i] != new_shape[i]:
                repeat_count = new_shape[i] # number of times to repeat data, coords
                cum_repeat *= repeat_count # update cumulative repeat factor
                nnz = len(new_data) # Number of non-zero elements so far

                # Tile data and coordinates to match the new repeat count
                new_data = np.tile(new_data, repeat_count)
                new_coords = tuple(np.tile(new_coords[i+1:], repeat_count))

                # Create new dimensions and stack them
                new_dim = np.repeat(np.arange(0, repeat_count, dtype=idx_dtype), nnz)
                new_coords = (new_dim,) + new_coords
            else:
                # If no broadcasting needed, tile the coordinates
                new_dim = np.tile(coords[i], cum_repeat)
                new_coords = (new_dim,) + new_coords

        return coo_array((new_data, new_coords), new_shape)

    def _sum_nd(self, axis, res_dtype, out):
        # axis and out are preprocessed. out.shape is new_shape
        A2d, new_shape = _convert_to_2d(self, axis)
        ones = np.ones((A2d.shape[1], 1), dtype=res_dtype)
        # sets dtype while loading into out
        out[...] = (A2d @ ones).reshape(new_shape)
        return out

    def _min_or_max_axis_nd(self, axis, min_or_max, explicit):
        A2d, new_shape = _convert_to_2d(self, axis)
        res = A2d._min_or_max_axis(1, min_or_max, explicit)
        unraveled_coords = np.unravel_index(res.coords[0], new_shape)

        return coo_array((res.data, unraveled_coords), new_shape)

    def _argminmax_axis_nd(self, axis, argminmax, compare, explicit):
        A2d, new_shape = _convert_to_2d(self, axis)
        res_flat = A2d._argminmax_axis(1, argminmax, compare, explicit)
        return res_flat.reshape(new_shape)


def _block_diag(self):
    """
    Converts an N-D COO array into a 2-D COO array in block diagonal form.

    Parameters:
    self (coo_array): An N-Dimensional COO sparse array.

    Returns:
    coo_array: A 2-Dimensional COO sparse array in block diagonal form.
    """
    if self.ndim<2:
        raise ValueError("array must have atleast dim=2")
    num_blocks = math.prod(self.shape[:-2])
    n_col = self.shape[-1]
    n_row = self.shape[-2]
    res_arr = self.reshape((num_blocks, n_row, n_col))
    new_coords = (
        res_arr.coords[1] + res_arr.coords[0] * res_arr.shape[1],
        res_arr.coords[2] + res_arr.coords[0] * res_arr.shape[2],
    )

    new_shape = (num_blocks * n_row, num_blocks * n_col)
    return coo_array((self.data, tuple(new_coords)), shape=new_shape)


def _extract_block_diag(self, shape):
    n_row, n_col = shape[-2], shape[-1]

    # Extract data and coordinates from the block diagonal COO array
    data = self.data
    row, col = self.row, self.col

    # Initialize new coordinates array
    new_coords = np.empty((len(shape), self.nnz), dtype=int)

    # Calculate within-block indices
    new_coords[-2] = row % n_row
    new_coords[-1] = col % n_col

    # Calculate coordinates for higher dimensions
    temp_block_idx = row // n_row
    for i in range(len(shape) - 3, -1, -1):
        size = shape[i]
        new_coords[i] = temp_block_idx % size
        temp_block_idx = temp_block_idx // size

    # Create the new COO array with the original n-D shape
    return coo_array((data, tuple(new_coords)), shape=shape)


def _get_sparse_data_and_coords(x, new_shape, dtype):
    x = x.tocoo()
    x.sum_duplicates()

    x_coords = list(x.coords)
    x_data = x.data.astype(dtype, copy=False)
    x_shape = x.shape

    if new_shape == x_shape:
        return x_data, x_coords

    # broadcasting needed
    len_diff = len(new_shape) - len(x_shape)
    if len_diff > 0:
        # prepend ones to shape of x to match ndim
        x_shape = [1] * len_diff + list(x_shape)
        coord_zeros = np.zeroslike(x_coords[0])
        x_coords = tuple([coord_zeros] * len_diff + x_coords)
    # taking away axes (squeezing) is not part of broadcasting, but long
    # spmatrix history of using 2d vectors in 1d space, so we manually
    # squeeze the front and back axes here to be compatible
    if len_diff < 0:
        for _ in range(-len_diff):
            if x_shape[0] == 1:
                x_shape = x_shape[1:]
                x_coords = x_coords[1:]
            elif x_shape[-1] == 1:
                x_shape = x_shape[:-1]
                x_coords = x_coords[:-1]
            else:
                raise ValueError("shape mismatch in assignment")
    # broadcast with copy (will need to copy eventually anyway)
    tot_expand = 1
    for i, (nn, nx) in enumerate(zip(new_shape, x_shape)):
        if nn == nx:
            continue
        if nx != 1:
            raise ValueError("shape mismatch in assignment")
        x_nnz = len(x_coords[0])
        x_coords[i] = np.repeat(np.arange(nn), x_nnz)
        for j, co in enumerate(x_coords):
            if j == i:
                continue
            x_coords[j] = np.tile(co, nn)
        tot_expand *= nn
    x_data = np.tile(x_data.ravel(), tot_expand)
    return x_data, x_coords


def _get_dense_data_and_coords(x, new_shape):
    if x.shape != new_shape:
        x = np.broadcast_to(x.squeeze(), new_shape)
    # shift scalar input to 1d so has coords
    if new_shape == ():
        x_coords = tuple([np.array([0])] * len(new_shape))
        x_data = x.ravel()
    else:
        x_coords = x.nonzero()
        x_data = x[x_coords]
    return x_data, x_coords


def _process_axes(ndim_a, ndim_b, axes):
    if isinstance(axes, int):
        if axes < 1 or axes > min(ndim_a, ndim_b):
            raise ValueError("axes integer is out of bounds for input arrays")
        axes_a = list(range(ndim_a - axes, ndim_a))
        axes_b = list(range(axes))
    elif isinstance(axes, tuple | list):
        if len(axes) != 2:
            raise ValueError("axes must be a tuple/list of length 2")
        axes_a, axes_b = axes
        if len(axes_a) != len(axes_b):
            raise ValueError("axes lists/tuples must be of the same length")
        if any(ax >= ndim_a or ax < -ndim_a for ax in axes_a) or \
           any(bx >= ndim_b or bx < -ndim_b for bx in axes_b):
            raise ValueError("axes indices are out of bounds for input arrays")
    else:
        raise TypeError("axes must be an integer or a tuple/list of integers")

    axes_a = [axis + ndim_a if axis < 0 else axis for axis in axes_a]
    axes_b = [axis + ndim_b if axis < 0 else axis for axis in axes_b]
    return axes_a, axes_b


def _convert_to_2d(coo, axis):
    axis_coords = tuple(coo.coords[i] for i in axis)
    axis_shape = tuple(coo.shape[i] for i in axis)
    axis_ravel = _ravel_coords(axis_coords, axis_shape)

    ndim = len(coo.coords)
    non_axis = tuple(i for i in range(ndim) if i not in axis)
    if non_axis:
        non_axis_coords = tuple(coo.coords[i] for i in non_axis)
        non_axis_shape = tuple(coo.shape[i] for i in non_axis)
        non_axis_ravel = _ravel_coords(non_axis_coords, non_axis_shape)
        coords_2d = (non_axis_ravel, axis_ravel)
        shape_2d = (math.prod(non_axis_shape), math.prod(axis_shape))
    else:  # all axes included in axis so result will have 1 element
        coords_2d = (axis_ravel,)
        shape_2d = (math.prod(axis_shape),)
        non_axis_shape = ()

    new_coo = coo_array((coo.data, coords_2d), shape=shape_2d)
    return new_coo, non_axis_shape


def _ravel_coords(coords, shape, order='C'):
    """Like np.ravel_multi_index, but avoids some overflow issues."""
    if len(coords) == 1:
        return coords[0]
    # Handle overflow as in https://github.com/scipy/scipy/pull/9132
    if len(coords) == 2:
        nrows, ncols = shape
        row, col = coords
        if order == 'C':
            maxval = (ncols * max(0, nrows - 1) + max(0, ncols - 1))
            idx_dtype = get_index_dtype(maxval=maxval)
            return np.multiply(ncols, row, dtype=idx_dtype) + col
        elif order == 'F':
            maxval = (nrows * max(0, ncols - 1) + max(0, nrows - 1))
            idx_dtype = get_index_dtype(maxval=maxval)
            return np.multiply(nrows, col, dtype=idx_dtype) + row
        else:
            raise ValueError("'order' must be 'C' or 'F'")
    return np.ravel_multi_index(coords, shape, order=order)


def isspmatrix_coo(x):
    """Is `x` of coo_matrix type?

    .. warning::

       SciPy sparse is shifting from a sparse matrix interface to a sparse
       array interface. In the next few releases we expect to deprecate the
       sparse matrix interface. For documentation of the matrix
       interface, see the :ref:`spmatrix interface docs <spmatrix_api>`.
       For guidance on converting existing code to sparse arrays, see
       :ref:`Migration from spmatrix to sparray <migration_to_sparray>`.

    Parameters
    ----------
    x
        object to check for being a coo matrix

    Returns
    -------
    bool
        True if `x` is a coo matrix, False otherwise

    Examples
    --------
    >>> from scipy.sparse import coo_array, coo_matrix, csr_matrix, isspmatrix_coo
    >>> isspmatrix_coo(coo_matrix([[5]]))
    True
    >>> isspmatrix_coo(coo_array([[5]]))
    False
    >>> isspmatrix_coo(csr_matrix([[5]]))
    False
    """
    return isinstance(x, coo_matrix)


# This namespace class separates array from matrix with isinstance
class coo_array(_coo_base, sparray):
    """
    A sparse array in COOrdinate format.

    Also known as the 'ijv' or 'triplet' format.

    This can be instantiated in several ways:
        coo_array(D)
            where D is an ndarray

        coo_array(S)
            with another sparse array or matrix S (equivalent to S.tocoo())

        coo_array(shape, [dtype])
            to construct an empty sparse array with shape `shape`
            dtype is optional, defaulting to dtype='d'.

        coo_array((data, coords), [shape])
            to construct from existing data and index arrays:
                1. data[:]       the entries of the sparse array, in any order
                2. coords[i][:]  the axis-i coordinates of the data entries

            Where ``A[coords] = data``, and coords is a tuple of index arrays.
            When shape is not specified, it is inferred from the index arrays.

    Attributes
    ----------
    data : ndarray
        COO format data array of the sparse array
    coords : tuple of ndarray
        COO format tuple of index arrays
    has_canonical_format : bool
        Whether the matrix has sorted coordinates and no duplicates
    dtype : dtype
        Data type of the array
    shape : tuple of integers
        Shape of the array
    ndim : int
        Number of dimensions of the array
    format : str
        Three letter code for the format of the array storage, e.g. 'coo'
    nnz : int
        Number of values stored in the array
    size : int
        Number of values stored in the array
    T : coo_array
        The transpose of the array
    mT : coo_array
        The matrix transpose of the array

    Notes
    -----

    Sparse arrays can be used in arithmetic operations: they support
    addition, subtraction, multiplication, division, and matrix power.

    Advantages of the COO format
        - facilitates fast conversion among sparse formats
        - permits duplicate entries (see example)
        - very fast conversion to and from CSR/CSC formats

    Disadvantages of the COO format
        - does not directly support:
            + arithmetic operations
            + slicing

    Intended Usage
        - COO is a fast format for constructing sparse arrays
        - Once a COO array has been constructed, convert to CSR or
          CSC format for fast arithmetic and matrix vector operations
        - By default when converting to CSR or CSC format, duplicate (i,j)
          entries will be summed together.  This facilitates efficient
          construction of finite element matrices and the like. (see example)

    Canonical format
        - Entries and coordinates sorted by row, then column.
        - There are no duplicate entries (i.e. duplicate (i,j) locations)
        - Data arrays MAY have explicit zeros.

    Examples
    --------

    >>> # Constructing an empty sparse array
    >>> import numpy as np
    >>> from scipy.sparse import coo_array
    >>> coo_array((3, 4), dtype=np.int8).toarray()
    array([[0, 0, 0, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 0]], dtype=int8)

    >>> # Constructing a sparse array using ijv format
    >>> row  = np.array([0, 3, 1, 0])
    >>> col  = np.array([0, 3, 1, 2])
    >>> data = np.array([4, 5, 7, 9])
    >>> coo_array((data, (row, col)), shape=(4, 4)).toarray()
    array([[4, 0, 9, 0],
           [0, 7, 0, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 5]])

    >>> # Constructing a sparse array with duplicate coordinates
    >>> row  = np.array([0, 0, 1, 3, 1, 0, 0])
    >>> col  = np.array([0, 2, 1, 3, 1, 0, 0])
    >>> data = np.array([1, 1, 1, 1, 1, 1, 1])
    >>> coo = coo_array((data, (row, col)), shape=(4, 4))
    >>> # Duplicate coordinates are maintained until implicitly or explicitly summed
    >>> np.max(coo.data)
    1
    >>> coo.toarray()
    array([[3, 0, 1, 0],
           [0, 2, 0, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 1]])

    """  # numpydoc ignore=PR01


class coo_matrix(spmatrix, _coo_base):
    """
    A sparse matrix in COOrdinate format.

    Also known as the 'ijv' or 'triplet' format.

    .. warning::

       SciPy sparse is shifting from a sparse matrix interface to a sparse
       array interface. In the next few releases we expect to deprecate the
       sparse matrix interface. For documentation of the matrix
       interface, see the :ref:`spmatrix interface docs <spmatrix_api>`.
       For guidance on converting existing code to sparse arrays, see
       :ref:`Migration from spmatrix to sparray <migration_to_sparray>`.

    This can be instantiated in several ways:
        coo_matrix(D)
            where D is a 2-D ndarray

        coo_matrix(S)
            with another sparse array or matrix S (equivalent to S.tocoo())

        coo_matrix((M, N), [dtype])
            to construct an empty matrix with shape (M, N)
            dtype is optional, defaulting to dtype='d'.

        coo_matrix((data, (i, j)), [shape=(M, N)])
            to construct from three arrays:
                1. data[:]   the entries of the matrix, in any order
                2. i[:]      the row indices of the matrix entries
                3. j[:]      the column indices of the matrix entries

            Where ``A[i[k], j[k]] = data[k]``.  When shape is not
            specified, it is inferred from the index arrays

    Attributes
    ----------
    data : ndarray
        COO format data array of the sparse matrix
    coords : tuple of ndarray
        COO format tuple of index matrix
    has_canonical_format : bool
        Whether the matrix has sorted coordinates and no duplicates
    dtype : dtype
        Data type of the matrix
    shape : tuple of integers
        Shape of the matrix
    ndim : int
        Number of dimensions of the matrix
    format : str
        Three letter code for the format of the matrix storage, e.g. 'coo'
    nnz : int
        Number of values stored in the matrix
    size : int
        Number of values stored in the matrix
    T : coo_matrix
        The transpose of the matrix
    mT : coo_matrix
        The matrix transpose

    Notes
    -----

    Sparse matrices can be used in arithmetic operations: they support
    addition, subtraction, multiplication, division, and matrix power.

    Advantages of the COO format
        - facilitates fast conversion among sparse formats
        - permits duplicate entries (see example)
        - very fast conversion to and from CSR/CSC formats

    Disadvantages of the COO format
        - does not directly support:
            + arithmetic operations
            + slicing

    Intended Usage
        - COO is a fast format for constructing sparse matrices
        - Once a COO matrix has been constructed, convert to CSR or
          CSC format for fast arithmetic and matrix vector operations
        - By default when converting to CSR or CSC format, duplicate (i,j)
          entries will be summed together.  This facilitates efficient
          construction of finite element matrices and the like. (see example)

    Canonical format
        - Entries and coordinates sorted by row, then column.
        - There are no duplicate entries (i.e. duplicate (i,j) locations)
        - Data arrays MAY have explicit zeros.

    Examples
    --------

    >>> # Constructing an empty matrix
    >>> import numpy as np
    >>> from scipy.sparse import coo_matrix
    >>> coo_matrix((3, 4), dtype=np.int8).toarray()
    array([[0, 0, 0, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 0]], dtype=int8)

    >>> # Constructing a matrix using ijv format
    >>> row  = np.array([0, 3, 1, 0])
    >>> col  = np.array([0, 3, 1, 2])
    >>> data = np.array([4, 5, 7, 9])
    >>> coo_matrix((data, (row, col)), shape=(4, 4)).toarray()
    array([[4, 0, 9, 0],
           [0, 7, 0, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 5]])

    >>> # Constructing a matrix with duplicate coordinates
    >>> row  = np.array([0, 0, 1, 3, 1, 0, 0])
    >>> col  = np.array([0, 2, 1, 3, 1, 0, 0])
    >>> data = np.array([1, 1, 1, 1, 1, 1, 1])
    >>> coo = coo_matrix((data, (row, col)), shape=(4, 4))
    >>> # Duplicate coordinates are maintained until implicitly or explicitly summed
    >>> np.max(coo.data)
    1
    >>> coo.toarray()
    array([[3, 0, 1, 0],
           [0, 2, 0, 0],
           [0, 0, 0, 0],
           [0, 0, 0, 1]])

    """

    def __setstate__(self, state):
        if 'coords' not in state:
            # For retro-compatibility with the previous attributes
            # storing nnz coordinates for 2D COO matrix.
            state['coords'] = (state.pop('row'), state.pop('col'))
        self.__dict__.update(state)

    def __getitem__(self, key):
        raise TypeError("'coo_matrix' object is not subscriptable")

    def __setitem__(self, key, x):
        raise TypeError("'coo_matrix' object does not support item assignment")



# file for distribution-specific tests with new infrastructure (UnivariateDistribution)
import numpy as np
from numpy.testing import assert_allclose
from scipy import stats

class TestBinomial:
    def test_gh23708_binomial_logcdf_method_complement(self):
        # gh-23708 found that `logcdf` method='complement' was inaccurate in the tails
        x = np.asarray([0., 18.])
        X = stats.Binomial(n=np.asarray([18.]), p=np.asarray(0.71022842))
        assert_allclose(X.logcdf(x, method='complement'), X.logcdf(x), rtol=1e-15)
        assert_allclose(X.logccdf(x, method='complement'), X.logccdf(x), rtol=1e-15)

        # going even deeper into the tails
        X = stats.Binomial(n=100, p=0.5)
        assert_allclose(X.logcdf(0, method='complement'), X.logpmf(0), rtol=1e-15)
        assert_allclose(X.logccdf(99, method='complement'), X.logpmf(100), rtol=1e-15)



import math
import numpy as np
from scipy.special import betainc
from scipy._lib._array_api import (
    xp_capabilities,
    xp_ravel,
    array_namespace,
    xp_promote,
    xp_device,
    xp_size,
    _count_nonmasked,
    is_torch,
    is_lazy_array,
)
import scipy._external.array_api_extra as xpx
from scipy.stats._axis_nan_policy import _broadcast_arrays, _contains_nan


def _quantile_iv(x, p, method, axis, nan_policy, keepdims, weights, fun='quantile'):
    xp = array_namespace(x, p, weights)

    if fun == "quantile":
        methods = {'inverted_cdf', 'averaged_inverted_cdf', 'closest_observation',
                   'hazen', 'interpolated_inverted_cdf', 'linear',
                   'median_unbiased', 'normal_unbiased', 'weibull',
                   'harrell-davis', '_lower', '_midpoint', '_higher', '_nearest',
                   'round_outward', 'round_inward', 'round_nearest'}
        allowed_types = 'real floating'
        def mask_fun(p): return (p > 1) | (p < 0) | xp.isnan(p)
        var2_name = 'p'
        var2_type_msg = '`p` must have real floating dtype.'
    else:
        methods = set(_estimated_cdf_methods)
        allowed_types = ('integral', 'real floating')
        mask_fun = xp.isnan
        var2_name = 'y'
        var2_type_msg = '`y` must have real dtype.'

    if not xp.isdtype(xp.asarray(x).dtype, ('integral', 'real floating')):
        raise ValueError("`x` must have real dtype.")

    if not xp.isdtype(xp.asarray(p).dtype, allowed_types):
        raise ValueError(var2_type_msg)

    if not (weights is None
            or xp.isdtype(xp.asarray(weights).dtype, ('integral', 'real floating'))):
        raise ValueError("`weights` must have real dtype.")

    x, p, weights = xp_promote(x, p, weights, force_floating=True, xp=xp)
    p = xp.asarray(p, device=xp_device(x))
    dtype = x.dtype

    axis_none = axis is None
    ndim = max(x.ndim, p.ndim)
    if axis_none:
        x = xp_ravel(x)
        p = xp_ravel(p)
        axis = 0
    elif np.iterable(axis) or int(axis) != axis:
        message = "`axis` must be an integer or None."
        raise ValueError(message)
    elif (axis >= ndim) or (axis < -ndim):
        message = "`axis` is not compatible with the shapes of the inputs."
        raise ValueError(message)
    axis = int(axis)

    if method not in methods:
        message = f"`method` must be one of {methods}"
        raise ValueError(message)

    no_weights = {'_lower', '_midpoint', '_higher', '_nearest', 'harrell-davis',
                  'round_nearest', 'round_inward', 'round_outward'}
    if weights is not None and method in no_weights:
        message = f"`method='{method}'` does not support `weights`."
        raise ValueError(message)

    contains_nans = _contains_nan(x, nan_policy, xp_omit_okay=True, xp=xp)

    if keepdims not in {None, True, False}:
        message = "If specified, `keepdims` must be True or False."
        raise ValueError(message)

    # If data has length zero along `axis`, the result will be an array of NaNs just
    # as if the data had length 1 along axis and were filled with NaNs. This is treated
    # naturally below whether `nan_policy` is `'propagate'` or `'omit'`.
    if x.shape[axis] == 0 and fun == 'quantile':
        shape = list(x.shape)
        shape[axis] = 1
        x = xp.full(shape, xp.nan, dtype=dtype, device=xp_device(x))

    if weights is None:
        y = xp.sort(x, axis=axis, stable=False)
        y, p = _broadcast_arrays((y, p), axis=axis)
        n_zero_weight = None
    else:
        x, weights = xp.broadcast_arrays(x, weights)
        i_zero_weight = (weights == 0)
        n_zero_weight = xp.count_nonzero(i_zero_weight, axis=axis, keepdims=True)
        x = xpx.at(x)[i_zero_weight].set(xp.inf, copy=True)
        i_y = xp.argsort(x, axis=axis, stable=False)
        y = xp.take_along_axis(x, i_y, axis=axis)
        weights = xp.take_along_axis(weights, i_y, axis=axis)
        y, p, weights, i_y, n_zero_weight = _broadcast_arrays(
            (y, p, weights, i_y, n_zero_weight), axis=axis)

    if (keepdims is False) and (p.shape[axis] != 1):
        message = ("`keepdims` may be False only if the length of "
                   f"`{var2_name}` along `axis` is 1.")
        raise ValueError(message)
    keepdims = (p.shape[axis] != 1) if keepdims is None else keepdims

    y = xp.moveaxis(y, axis, -1)
    p = xp.moveaxis(p, axis, -1)
    weights = weights if weights is None else xp.moveaxis(weights, axis, -1)
    n_zero_weight = (n_zero_weight if n_zero_weight is None
                     else xp.moveaxis(n_zero_weight, axis, -1))

    n = _count_nonmasked(y, -1, xp=xp, keepdims=True)
    n = n if n_zero_weight is None else n - n_zero_weight

    nan_out = None
    if is_lazy_array(y) or contains_nans:
        nans = xp.isnan(y)

        # Note that if length along `axis` were 0 to begin with,
        # it is now length 1 and filled with NaNs.
        if nan_policy == 'propagate':
            nan_out = xp.any(nans, axis=-1, keepdims=True)
        else:  # 'omit'
            n_int = n - xp.count_nonzero(nans, axis=-1, keepdims=True)
            n = xp.astype(n_int, dtype)
            # NaNs are produced only if slice is empty after removing NaNs
            nan_out = n == 0
            n = xpx.at(n, nan_out).set(y.shape[-1])  # avoids pytorch/pytorch#146211

        if (is_lazy_array(nans) or xp.any(nans)) and (method == 'harrell-davis'):
            y = xp.asarray(y, copy=True)  # ensure writable
            y = xpx.at(y, nans).set(0)  # any non-nan will prevent NaN from propagating
        if is_lazy_array(nan_out) or xp.any(nan_out):
            y = xp.asarray(y, copy=True)  # ensure writable
            y = xpx.at(y, xp.broadcast_to(nan_out, y.shape)).set(xp.nan)

    n = xp.asarray(n, dtype=dtype, device=xp_device(y))

    # apparently xpx.at is accepting nan_out as a mask even though it doesn't have the
    # same number of dimensions as `y`, yet it still appears to work correctly?
    # should refactor for clarity, and p_mask that gets returned here should probably
    # get a different name - `nan_out` would be more appropriate!
    p_mask = mask_fun(p) if nan_out is None else mask_fun(p) | nan_out
    if is_lazy_array(p_mask) or xp.any(p_mask):
        p = xp.where(p_mask, 0.5, p)  # these get NaN-ed out at the end

    return (y, p, method, axis, nan_policy, keepdims,
            n, axis_none, ndim, p_mask, weights, xp)


@xp_capabilities(skip_backends=[("dask.array", "No take_along_axis yet.")],
                 marray=True,
                 extra_note=("Use of `weights` is incompatible with MArray. "
                             "``method='harrell-davis'`` is CPU-only."))
def quantile(x, p, *, method='linear', axis=0, nan_policy='propagate', keepdims=None,
             weights=None):
    """
    Compute the p-th quantile of the data along the specified axis.

    Parameters
    ----------
    x : array_like of real numbers
        Data array.
    p : array_like of float
        Probability or sequence of probabilities of the quantiles to compute.
        Values must be between 0 and 1 (inclusive).
        While `numpy.quantile` can only compute quantiles according to the Cartesian
        product of the first two arguments, this function enables calculation of
        quantiles at different probabilities for each axis slice by following
        broadcasting rules like those of `scipy.stats` reducing functions.
        See `axis`, `keepdims`, and the examples.
    method : str, default: 'linear'
        The method to use for estimating the quantile.
        The available options, numbered as they appear in [1]_, are:

        1. 'inverted_cdf'
        2. 'averaged_inverted_cdf'
        3. 'closest_observation'
        4. 'interpolated_inverted_cdf'
        5. 'hazen'
        6. 'weibull'
        7. 'linear'  (default)
        8. 'median_unbiased'
        9. 'normal_unbiased'

        'harrell-davis' is also available to compute the quantile estimate
        according to [2]_.

        'round_outward', 'round_inward', and 'round_nearest' are available for use
        in trimming and winsorizing data.

        See Notes for details.
    axis : int or None, default: 0
        Axis along which the quantiles are computed.
        ``None`` ravels both `x` and `p` before performing the calculation,
        without checking whether the original shapes were compatible.
        As in other `scipy.stats` functions, a positive integer `axis` is resolved
        after prepending 1s to the shape of `x` or `p` as needed until the two arrays
        have the same dimensionality. When providing `x` and `p` with different
        dimensionality, consider using negative `axis` integers for clarity.
    nan_policy : str, default: 'propagate'
        Defines how to handle NaNs in the input data `x`.

        - ``propagate``: if a NaN is present in the axis slice (e.g. row) along
          which the  statistic is computed, the corresponding slice of the output
          will contain NaN(s).
        - ``omit``: NaNs will be omitted when performing the calculation.
          If insufficient data remains in the axis slice along which the
          statistic is computed, the corresponding slice of the output will
          contain NaN(s).
        - ``raise``: if a NaN is present, a ``ValueError`` will be raised.

        If NaNs are present in `p`, a ``ValueError`` will be raised.
    keepdims : bool, optional
        Consider the case in which `x` is 1-D and `p` is a scalar: the quantile
        is a reducing statistic, and the default behavior is to return a scalar.
        If `keepdims` is set to True, the axis will not be reduced away, and the
        result will be a 1-D array with one element.

        The general case is more subtle, since multiple quantiles may be
        requested for each axis-slice of `x`. For instance, if both `x` and `p`
        are 1-D and ``p.size > 1``, no axis can be reduced away; there must be an
        axis to contain the number of quantiles given by ``p.size``. Therefore:

        - By default, the axis will be reduced away if possible (i.e. if there is
          exactly one element of `p` per axis-slice of `x`).
        - If `keepdims` is set to True, the axis will not be reduced away.
        - If `keepdims` is set to False, the axis will be reduced away
          if possible, and an error will be raised otherwise.

    weights : array_like of finite, non-negative real numbers
        Frequency weights; e.g., for counting number weights,
        ``quantile(x, p, weights=weights)`` is equivalent to
        ``quantile(np.repeat(x, weights), p)``. Values other than finite counting
        numbers are accepted, but may not have valid statistical interpretations.
        Not compatible with ``method='harrell-davis'`` or those that begin with
        ``'round_'``.

    Returns
    -------
    quantile : scalar or ndarray
        The resulting quantile(s). The dtype is the result dtype of `x` and `p`.

    See Also
    --------
    numpy.quantile
    estimated_cdf
    :ref:`outliers`

    Notes
    -----
    Given a sample `x` from an underlying distribution, `quantile` provides a
    nonparametric estimate of the inverse cumulative distribution function.

    By default, this is done by interpolating between adjacent elements in
    ``y``, a sorted copy of `x`::

        (1-g)*y[j] + g*y[j+1]

    where the index ``j`` and coefficient ``g`` are the integral and
    fractional components of ``p * (n-1)``, and ``n`` is the number of
    elements in the sample.

    This is a special case of Equation 1 of H&F [1]_. More generally,

    - ``j = (p*n + m - 1) // 1``, and
    - ``g = (p*n + m - 1) % 1``,

    where ``m`` may be defined according to several different conventions.
    The preferred convention may be selected using the ``method`` parameter:

    =============================== =============== ===============
    ``method``                      number in H&F   ``m``
    =============================== =============== ===============
    ``interpolated_inverted_cdf``   4               ``0``
    ``hazen``                       5               ``1/2``
    ``weibull``                     6               ``p``
    ``linear`` (default)            7               ``1 - p``
    ``median_unbiased``             8               ``p/3 + 1/3``
    ``normal_unbiased``             9               ``p/4 + 3/8``
    =============================== =============== ===============

    Note that indices ``j`` and ``j + 1`` are clipped to the range ``0`` to
    ``n - 1`` when the results of the formula would be outside the allowed
    range of non-negative indices. When ``j`` is clipped to zero, ``g`` is
    set to zero as well. The ``-1`` in the formulas for ``j`` and ``g``
    accounts for Python's 0-based indexing.

    The table above includes only the estimators from [1]_ that are continuous
    functions of probability `p` (estimators 4-9). SciPy also provides the
    three discontinuous estimators from [1]_ (estimators 1-3), where ``j`` is
    defined as above, ``m`` is defined as follows, and ``g`` is ``0`` when
    ``index = p*n + m - 1`` is less than ``0`` and otherwise is defined below.

    1. ``inverted_cdf``: ``m = 0`` and ``g = int(index - j > 0)``
    2. ``averaged_inverted_cdf``: ``m = 0`` and
       ``g = (1 + int(index - j > 0)) / 2``
    3. ``closest_observation``: ``m = -1/2`` and
       ``g = 1 - int((index == j) & (j%2 == 1))``

    Note that for methods ``inverted_cdf`` and ``averaged_inverted_cdf``, only the
    relative proportions of tied observations (and relative weights) affect the
    results; for all other methods, the total number of observations (and absolute
    weights) matter.

    A different strategy for computing quantiles from [2]_, ``method='harrell-davis'``,
    uses a weighted combination of all elements. The weights are computed as:

    .. math::

        w_{n, i} = I_{i/n}(a, b) - I_{(i - 1)/n}(a, b)

    where :math:`n` is the number of elements in the sample,
    :math:`i` are the indices :math:`1, 2, ..., n-1, n` of the sorted elements,
    :math:`a = p (n + 1)`, :math:`b = (1 - p)(n + 1)`,
    :math:`p` is the probability of the quantile, and
    :math:`I` is the regularized, lower incomplete beta function
    (`scipy.special.betainc`).

    ``method='round_nearest'`` is equivalent to indexing ``y[j]``, where::

        j = int(np.round(p*n) if p < 0.5 else np.round(n*p - 1))

    This is useful when winsorizing data: replacing ``p*n`` of the most extreme
    observations with the next most extreme observation. ``method='round_outward'``
    adjusts the direction of rounding to winsorize fewer elements::

        j = int(np.floor(p*n) if p < 0.5 else np.ceil(n*p - 1))

    and ``method='round_inward'`` rounds to winsorize more elements::

        j = int(np.ceil(p*n) if p < 0.5 else np.floor(n*p - 1))

    These methods are also useful for trimming data: removing ``p*n`` of the most
    extreme observations. See :ref:`outliers` for example applications.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy import stats
    >>> x = np.asarray([[10, 8, 7, 5, 4],
    ...                 [0, 1, 2, 3, 5]])

    Take the median of each row.

    >>> stats.quantile(x, 0.5, axis=-1)
    array([7.,  2.])

    Take a different quantile for each row.

    >>> stats.quantile(x, [[0.25], [0.75]], axis=-1, keepdims=True)
    array([[5.],
           [3.]])

    Take multiple quantiles for each row.

    >>> stats.quantile(x, [0.25, 0.75], axis=-1)
    array([[5., 8.],
           [1., 3.]])

    Take different quantiles for each row.

    >>> p = np.asarray([[0.25, 0.75],
    ...                 [0.5, 1.0]])
    >>> stats.quantile(x, p, axis=-1)
    array([[5., 8.],
           [2., 5.]])

    Take different quantiles for each column.

    >>> stats.quantile(x.T, p.T, axis=0)
    array([[5., 2.],
           [8., 5.]])

    References
    ----------
    .. [1] R. J. Hyndman and Y. Fan,
       "Sample quantiles in statistical packages,"
       The American Statistician, 50(4), pp. 361-365, 1996
    .. [2] Harrell, Frank E., and C. E. Davis.
       "A new distribution-free quantile estimator."
       Biometrika 69.3 (1982): 635-640.

    """
    # Input validation / standardization

    temp = _quantile_iv(x, p, method, axis, nan_policy, keepdims, weights)
    (y, p, method, axis, nan_policy, keepdims,
     n, axis_none, ndim, p_mask, weights, xp) = temp

    if method in {'inverted_cdf', 'averaged_inverted_cdf', 'closest_observation',
                  'hazen', 'interpolated_inverted_cdf', 'linear',
                  'median_unbiased', 'normal_unbiased', 'weibull'}:
        res = _quantile_hf(y, p, n, method, weights, xp)
    elif method in {'harrell-davis'}:
        res = _quantile_hd(y, p, n, xp)
    elif method in {'_lower', '_midpoint', '_higher', '_nearest'}:
        res = _quantile_bc(y, p, n, method, xp)
    else:  # method.startswith('round'):
        res = _quantile_winsor(y, p, n, method, xp)

    return _post_quantile(res, p_mask, axis, axis_none, ndim, keepdims, xp)


def _post_quantile(res, p_mask, axis, axis_none, ndim, keepdims, xp):
    res = xpx.at(res, p_mask).set(xp.nan)

    # Reshape per axis/keepdims
    if axis_none and keepdims:
        shape = (1,)*(ndim - 1) + res.shape
        res = xp.reshape(res, shape)
        axis = -1

    res = xp.moveaxis(res, -1, axis)

    if not keepdims:
        res = xp.squeeze(res, axis=axis)

    return res[()] if res.ndim == 0 else res


def _quantile_hf(y, p, n, method, weights, xp):
    ms = dict(inverted_cdf=0, averaged_inverted_cdf=0, closest_observation=-0.5,
              interpolated_inverted_cdf=0, hazen=0.5, weibull=p, linear=1 - p,
              median_unbiased=p/3 + 1/3, normal_unbiased=p/4 + 3/8)
    m = ms[method]

    if weights is None:
        jg = p * n + m
        jp1 = jg // 1
        j = jp1 - 1
    else:
        cumulative_weights = xp.cumulative_sum(weights, axis=-1)
        n_int = xp.asarray(n, dtype=xp.int64)
        n_int = xp.broadcast_to(n_int, cumulative_weights.shape[:-1] + (1,))
        total_weight = xp.take_along_axis(cumulative_weights, n_int-1, axis=-1)
        jg = p * total_weight + m
        jp1 = _xp_searchsorted(cumulative_weights, jg, side='right')
        j = _xp_searchsorted(cumulative_weights, jg-1, side='right')
        j, jp1 = xp.astype(j, y.dtype), xp.astype(jp1, y.dtype)

    g = jg % 1
    if method == 'inverted_cdf':
        g = xp.astype((g > 0), jg.dtype)
    elif method == 'averaged_inverted_cdf':
        g = (1 + xp.astype((g > 0), jg.dtype)) / 2
    elif method == 'closest_observation':
        g = (1 - xp.astype((g == 0) & (j % 2 == 1), jg.dtype))
    if method in {'inverted_cdf', 'averaged_inverted_cdf', 'closest_observation'}:
        g = xp.asarray(g)
        g = xpx.at(g, jg < 0).set(0)

    g = xpx.at(g)[j < 0].set(0)
    j = xp.clip(j, 0., n - 1)
    jp1 = xp.clip(jp1, 0., n - 1)

    return ((1 - g) * xp.take_along_axis(y, xp.astype(j, xp.int64), axis=-1)
            + g * xp.take_along_axis(y, xp.astype(jp1, xp.int64), axis=-1))


def _quantile_hd(y, p, n, xp):
    # RE axis handling: We need to perform a reducing operation over rows of `y` for
    # each element in the corresponding row of `p` (a la Cartesian product). Strategy:
    # move rows of `p` to an axis at the front that is orthogonal to all the rest,
    # perform the reducing operating over the last axis, then move the front axis back
    # to the end.
    p = xp.moveaxis(p, -1, 0)[..., xp.newaxis]
    a = p * (n + 1)
    b = (1 - p) * (n + 1)
    i = xp.arange(y.shape[-1] + 1, dtype=y.dtype, device=xp_device(y))
    w = betainc(a, b, i / n)
    w = w[..., 1:] - w[..., :-1]
    w = xpx.at(w, xp.isnan(w)).set(0)
    res = xp.vecdot(w, y, axis=-1)
    return xp.moveaxis(res, 0, -1)


def _quantile_winsor(y, p, n, method, xp):
    ops = dict(round_outward=(xp.floor, xp.ceil),
               round_inward=(xp.ceil, xp.floor),
               round_nearest=(xp.round, xp.round))
    op_left, op_right = ops[method]
    j = xp.where(p < 0.5, op_left(p*n), op_right(n*p - 1))
    return xp.take_along_axis(y, xp.astype(j, xp.int64), axis=-1)


def _quantile_bc(y, p, n, method, xp):
    # Methods retained for backward compatibility. NumPy documentation is not
    # quite right about what these methods do: if `p * (n - 1)` is integral,
    # that is used as the index. See numpy/numpy#28910.
    ij = p * (n - 1)
    if method == '_midpoint':
        return (xp.take_along_axis(y, xp.astype(xp.floor(ij), xp.int64), axis=-1)
                + xp.take_along_axis(y, xp.astype(xp.ceil(ij), xp.int64), axis=-1)) / 2
    elif method == '_lower':
        k = xp.floor(ij)
    elif method == '_higher':
        k = xp.ceil(ij)
    elif method == '_nearest':
        k = xp.round(ij)
    return xp.take_along_axis(y, xp.astype(k, xp.int64), axis=-1)


@xp_capabilities(skip_backends=[("dask.array", "No take_along_axis yet.")])
def _xp_searchsorted(x, y, *, side='left', xp=None):
    # Vectorized xp.searchsorted. Search is performed along last axis. The shape of the
    # output is that of `y`, broadcasting the batch dimensions with those of `x` if
    # necessary.
    xp = array_namespace(x, y) if xp is None else xp
    xp_default_int = xp.asarray(1).dtype
    y_0d = xp.asarray(y).ndim == 0
    x, y = _broadcast_arrays((x, y), axis=-1, xp=xp)
    x_1d = x.ndim <= 1

    if x_1d or is_torch(xp):
        y = xp.reshape(y, ()) if (y_0d and x_1d) else y
        out = xp.searchsorted(x, y, side=side)
        out = xp.astype(out, xp_default_int, copy=False)
        return out

    a = xp.full(y.shape, 0, device=xp_device(x))

    if x.shape[-1] == 0:
        return a

    n = xp.count_nonzero(~xp.isnan(x), axis=-1, keepdims=True)
    b = xp.broadcast_to(n, y.shape)

    compare = xp.less_equal if side == 'left' else xp.less

    # while xp.any(b - a > 1):
    # refactored to for loop with ~log2(n) iterations for JAX JIT
    for i in range(int(math.log2(x.shape[-1])) + 1):
        c = (a + b) // 2
        x0 = xp.take_along_axis(x, c, axis=-1)
        j = compare(y, x0)
        b = xp.where(j, c, b)
        a = xp.where(j, a, c)

    out = xp.where(compare(y, xp.min(x, axis=-1, keepdims=True)), 0, b)
    out = xp.where(xp.isnan(y), x.shape[-1], out) if side == 'right' else out
    out = xp.astype(out, xp_default_int, copy=False)
    return out


@xp_capabilities(skip_backends=[("dask.array", "No take_along_axis yet.")],
                 marray=True)
def estimated_cdf(x, y, *, method='linear',
                  axis=0, nan_policy='propagate', keepdims=None):
    """
    Estimate the CDF of the distribution underlying a sample.

    Parameters
    ----------
    x : array_like of real numbers
        Data array.
    y : array_like of real numbers
        Points at which to evaluate the estimated cdf.
        See `axis`, `keepdims`, and the examples for broadcasting behavior.
    method : str, default: 'linear'
        The method to use for estimating the cumulative distribution function.
        The available options, numbered as they appear in [1]_, are:

        1. 'inverted_cdf' (AKA *the* empirical cumulative distribution function)
        2. 'averaged_inverted_cdf'
        3. 'closest_observation'
        4. 'interpolated_inverted_cdf'
        5. 'hazen'
        6. 'weibull'
        7. 'linear'  (default)
        8. 'median_unbiased'
        9. 'normal_unbiased'

        See Notes for details.
    axis : int or None, default: 0
        Axis along which samples in `x` are given in ND case.
        ``None`` ravels both `x` and `y` before performing the calculation,
        without checking whether the original shapes were compatible.
        As in other `scipy.stats` functions, a positive integer `axis` is resolved
        after prepending 1s to the shape of `x` or `y` as needed until the two arrays
        have the same dimensionality. When providing `x` and `y` with different
        dimensionality, consider using negative `axis` integers for clarity.
    nan_policy : str, default: 'propagate'
        Defines how to handle NaNs in the input data `x`.

        - ``propagate``: if a NaN is present in the axis slice (e.g. row) along
          which the  statistic is computed, the corresponding slice of the output
          will contain NaN(s).
        - ``omit``: NaNs will be omitted when performing the calculation.
          If insufficient data remains in the axis slice along which the
          statistic is computed, the corresponding slice of the output will
          contain NaN(s).
        - ``raise``: if a NaN is present, a ``ValueError`` will be raised.

        If NaNs are present in `y`, the corresponding entries in the output will be NaN.
    keepdims : bool, optional
        Consider the case in which `x` is 1-D and `y` is a scalar: the estimated
        cumulative distribution function at a point is a reducing statistic, and the
        default behavior is to return a scalar.
        If `keepdims` is set to True, the axis will not be reduced away, and the
        result will be a 1-D array with one element.

        The general case is more subtle, since multiple values of `y` may be
        specified for each axis-slice of `x`. For instance, if both `x` and `y`
        are 1-D and ``y.size > 1``, no axis can be reduced away; there must be an
        axis to contain the number of values given by ``y.size``. Therefore:

        - By default, the axis will be reduced away if possible (i.e. if there is
          exactly one element of `y` per axis-slice of `x`).
        - If `keepdims` is set to True, the axis will not be reduced away.
        - If `keepdims` is set to False, the axis will be reduced away
          if possible, and an error will be raised otherwise.

    Returns
    -------
    probability : scalar or ndarray
        The resulting probabilities(s). The dtype is the result dtype of `x`, `y`, and
        the Python ``float`` type.

    See Also
    --------
    quantile

    Notes
    -----
    Given a sample `x` from an underlying distribution, `estimated_cdf` provides a
    nonparametric estimate of the cumulative distribution function.

    By default, this is done by computing the "fractional index" ``p`` at which ``y``
    would appear within ``z``, a sorted copy of `x`::

        p = 1 / (n - 1) * (j +  (     y - z[j])
                              / (z[j+1] - z[j]))

    where the index ``j`` is that of the largest element of ``z`` that does not exceed
    ``y``, and ``n`` is the number of elements in the sample. Note that if ``y`` is an
    element of ``z``, then ``j`` is the index such that ``y = z[j]``, and the formula
    simplifies to the intuitive ``j / (n - 1)``. The full formula linearly interpolates
    between ``j / (n - 1)`` and ``(j + 1) / (n - 1)``.

    This is a special case of the more general::

        p = (j + (y - z[j]) / (z[j+1] - z[j] + 1 - m) / n

    where ``m`` may be defined according to several different conventions.
    The preferred convention may be selected using the ``method`` parameter:

    =============================== =============== ===============
    ``method``                      number in H&F   ``m``
    =============================== =============== ===============
    ``interpolated_inverted_cdf``   4               ``0``
    ``hazen``                       5               ``1/2``
    ``weibull``                     6               ``p``
    ``linear`` (default)            7               ``1 - p``
    ``median_unbiased``             8               ``p/3 + 1/3``
    ``normal_unbiased``             9               ``p/4 + 3/8``
    =============================== =============== ===============

    Note that indices ``j`` and ``j + 1`` are clipped to the range ``0`` to
    ``n - 1`` when the results of the formula would be outside the allowed
    range of non-negative indices, and resulting ``p`` is clipped to the range
    ``0`` to ``1``.

    The table above includes only the estimators from [1]_ that are continuous
    functions of probability `p` (estimators 4-9). SciPy also provides the
    three discontinuous estimators from [1]_ (estimators 1-3), where ``j`` is
    defined as above, ``m`` is defined as follows, and ``g`` is ``0`` when
    ``index = p*n + m - 1`` is less than ``0`` and otherwise is defined below.

    1. ``inverted_cdf``: ``m = 0`` and ``g = int(index - j > 0)``
    2. ``averaged_inverted_cdf``: ``m = 0`` and
       ``g = (1 + int(index - j > 0)) / 2``
    3. ``closest_observation``: ``m = -1/2`` and
       ``g = 1 - int((index == j) & (j%2 == 1))``

    When all the data in ``x`` are unique, `estimated_cdf` and `quantile` are are
    inverses of one another within a certain domain.
    Although `quantile` with ``method='linear'`` is invertible over the whole domain
    of ``p`` from ``0`` to ``1``, this is not true of other methods.

    References
    ----------
    .. [1] R. J. Hyndman and Y. Fan,
           "Sample quantiles in statistical packages,"
           The American Statistician, 50(4), pp. 361-365, 1996

    Examples
    --------
    >>> import numpy as np
    >>> from scipy import stats
    >>> x = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    Estimate the cumulative distribution function of one sample at a single point.

    >>> stats.estimated_cdf(x, 5, axis=-1)
    np.float64(0.5)

    Estimate the cumulative distribution function of one sample at two points.

    >>> stats.estimated_cdf(x, [2.5, 7.5], axis=-1)
    array([0.25, 0.75])

    Estimate the cumulative distribution function of two samples at different points.

    >>> x = np.stack((np.arange(0, 11), np.arange(10, 21)))
    >>> stats.estimated_cdf(x, [[2.5], [17.5]], axis=-1, keepdims=True)
    array([[0.25],
           [0.75]])

    Estimate the cumulative distribution function at many points for each of two
    samples.

    >>> import matplotlib.pyplot as plt
    >>> rng = np.random.default_rng(6110515095)
    >>> x = stats.Normal(mu=[-1, 1]).sample(10000)
    >>> y = np.linspace(-4, 4, 5000)[:, np.newaxis]
    >>> p = stats.estimated_cdf(x, y, axis=0)
    >>> plt.plot(y, p)
    >>> plt.show()

    Note that the `quantile` and `estimated_cdf` functions are inverses of one another
    within a certain domain.

    >>> p = np.linspace(0, 1, 300)
    >>> x = rng.standard_normal(300)
    >>> y = stats.quantile(x, p)
    >>> p2 = stats.estimated_cdf(x, y)
    >>> np.testing.assert_allclose(p2, p)
    >>> y2 = stats.quantile(x, p2)
    >>> np.testing.assert_allclose(y2, y)

    However, the domain over which `quantile` can be inverted by `estimated_cdf` depends
    on the `method` used. This is most noticeable when there are few observations in the
    sample.

    >>> x = np.asarray([0, 1])
    >>> y_linear = stats.quantile(x, p, method='linear')
    >>> y_weibull = stats.quantile(x, p, method='weibull')
    >>> y_iicdf = stats.quantile(x, p, method='interpolated_inverted_cdf')
    >>> plt.plot(p, y_linear, p, y_weibull, p, y_iicdf)
    >>> plt.legend(['linear', 'weibull', 'iicdf'])
    >>> plt.xlabel('p')
    >>> plt.ylabel('y = quantile(x, p)')
    >>> plt.show()

    For example, in the case above, `quantile` is only invertible from
    ``p = 0.5`` to ``p = 1.0`` with ``method = 'interpolated_inverted_cdf'``. This is a
    fundamental characteristic of the methods, not a shortcoming of `estimated_cdf`.

    """
    temp = _quantile_iv(x, y, method, axis, nan_policy, keepdims, weights=None,
                        fun='estimated_cdf')
    x, y, method, axis, nan_policy, keepdims, n, axis_none, ndim, y_mask, _, xp = temp

    if xp_size(x) == 0:
        res = xp.full_like(y, xp.nan)
    else:
        res = _estimated_cdf_hf(x, y, n, method, xp)

    return _post_quantile(res, y_mask, axis, axis_none, ndim, keepdims, xp)


_estimated_cdf_discontinuous_methods = dict(
    inverted_cdf=0.0,
    averaged_inverted_cdf=0.0,
    closest_observation=0.5,
)


_estimated_cdf_continuous_methods = dict(
    interpolated_inverted_cdf=(0, 1),
    hazen=(0.5, 0.5),
    weibull=(0, 0),
    linear=(1, 1),
    median_unbiased=(1 / 3, 1 / 3),
    normal_unbiased=(3 / 8, 3 / 8),
)


_estimated_cdf_methods = (set(_estimated_cdf_continuous_methods).union(
                          set(_estimated_cdf_discontinuous_methods)))


def _estimated_cdf_hf(x, y, n, method, xp):
    n_int = xp.astype(n, xp.int64)
    j_max = n_int - 1
    j_min = xp.minimum(j_max, xp.asarray(1, dtype=j_max.dtype))
    jp1 = _xp_searchsorted(x, y, side='right')

    if method in _estimated_cdf_discontinuous_methods:
        dp = _estimated_cdf_discontinuous_methods[method]
        p = (xp.astype(jp1, x.dtype)+dp)/n

    else:
        jp1 = xp.clip(jp1, j_min, j_max)
        j = xp.clip(jp1-1, 0)
        xj = xp.take_along_axis(x, j, axis=-1)
        xjp1 = xp.take_along_axis(x, jp1, axis=-1)
        with np.errstate(divide='ignore', invalid='ignore'):  # refactor to apply_where?
            delta = xp.where((xjp1 > xj) & xp.isfinite(xj), (y - xj) / (xjp1 - xj), 1.)

        a, b = _estimated_cdf_continuous_methods[method]
        p = (xp.astype(jp1, x.dtype) + delta - a) / (n + 1 - a - b)

    xmin = x[..., :1]
    xmax = (x[..., -1:] if n.shape == () else xp.take_along_axis(x, j_max))
    p = xpx.at(p)[y < xmin].set(0.)
    p = xpx.at(p)[y > xmax].set(1.)

    return xp.clip(p, 0., 1.)



import numpy as np
import scipy._external.array_api_extra as xpx
from scipy._lib._array_api import xp_capabilities, array_namespace, xp_promote, is_jax
from scipy.optimize.elementwise import find_root
from scipy.special import ndtri
from scipy.special import _ufuncs as scu
from ._common import ConfidenceInterval


class BinomTestResult:
    """
    Result of `scipy.stats.binomtest`.

    Attributes
    ----------
    k : int
        The number of successes (copied from `binomtest` input).
    n : int
        The number of trials (copied from `binomtest` input).
    alternative : str
        Indicates the alternative hypothesis specified in the input
        to `binomtest`.  It will be one of ``'two-sided'``, ``'greater'``,
        or ``'less'``.
    statistic: float
        The estimate of the proportion of successes.
    pvalue : float
        The p-value of the hypothesis test.

    """
    def __init__(self, k, n, alternative, statistic, pvalue, xp):
        self.k = k
        self.n = n
        self.alternative = alternative
        self.statistic = statistic
        self.pvalue = pvalue
        self._xp = xp

        # add alias for backward compatibility
        self.proportion_estimate = statistic

    def __repr__(self):
        s = ("BinomTestResult("
             f"k={self.k}, "
             f"n={self.n}, "
             f"alternative={self.alternative!r}, "
             f"statistic={self.statistic}, "
             f"pvalue={self.pvalue})")
        return s

    def proportion_ci(self, confidence_level=0.95, method='exact'):
        """
        Compute the confidence interval for ``statistic``.

        Parameters
        ----------
        confidence_level : float, optional
            Confidence level for the computed confidence interval
            of the estimated proportion. Default is 0.95.
        method : {'exact', 'wilson', 'wilsoncc'}, optional
            Selects the method used to compute the confidence interval
            for the estimate of the proportion:

            'exact' :
                Use the Clopper-Pearson exact method [1]_.
            'wilson' :
                Wilson's method, without continuity correction ([2]_, [3]_).
            'wilsoncc' :
                Wilson's method, with continuity correction ([2]_, [3]_).

            Default is ``'exact'``.

        Returns
        -------
        ci : ``ConfidenceInterval`` object
            The object has attributes ``low`` and ``high`` that hold the
            lower and upper bounds of the confidence interval.

        References
        ----------
        .. [1] C. J. Clopper and E. S. Pearson, The use of confidence or
               fiducial limits illustrated in the case of the binomial,
               Biometrika, Vol. 26, No. 4, pp 404-413 (Dec. 1934).
        .. [2] E. B. Wilson, Probable inference, the law of succession, and
               statistical inference, J. Amer. Stat. Assoc., 22, pp 209-212
               (1927).
        .. [3] Robert G. Newcombe, Two-sided confidence intervals for the
               single proportion: comparison of seven methods, Statistics
               in Medicine, 17, pp 857-872 (1998).

        Examples
        --------
        >>> from scipy.stats import binomtest
        >>> result = binomtest(k=7, n=50, p=0.1)
        >>> result.statistic
        0.14
        >>> result.proportion_ci()
        ConfidenceInterval(low=0.05819170033997342, high=0.26739600249700846)
        """
        if method not in ('exact', 'wilson', 'wilsoncc'):
            raise ValueError(f"method ('{method}') must be one of 'exact', "
                             "'wilson' or 'wilsoncc'.")
        if not (0 <= confidence_level <= 1):
            raise ValueError(f'confidence_level ({confidence_level}) must be in '
                             'the interval [0, 1].')

        xp = self._xp
        k, n, confidence_level = xp_promote(self.k, self.n, confidence_level, xp=xp)
        if method == 'exact':
            low, high = _binom_exact_conf_int(
                k, n, confidence_level, self.alternative, xp=xp)
        else:
            # method is 'wilson' or 'wilsoncc'
            correction = method == 'wilsoncc'
            low, high = _binom_wilson_conf_int(
                k, n, confidence_level, self.alternative, correction, xp=xp)
        low = low[()] if low.ndim == 0 else low
        high = high[()] if high.ndim == 0 else high
        return ConfidenceInterval(low=low, high=high)


def _binom_exact_conf_int(k, n, confidence_level, alternative, *, xp):
    """
    Compute the estimate and confidence interval for the binomial test.

    Returns proportion, prop_low, prop_high
    """
    init = (xp.zeros_like(k), xp.ones_like(k))
    args = (k, n)
    alpha = ((1 - confidence_level) / 2 if alternative == 'two-sided'
             else 1 - confidence_level)

    # I think using the private methods here is fine, since we will only evaluate with
    # valid `p`, `k`, and `n` (or all NaNs). One exception is when `k=0` and
    # `binom._sf(k-1, n, p)`: evaluates to NaN, but that's not a problem because
    # `plow` has a special case for `k=0` below.
    plow = (xp.zeros_like(k) if alternative == 'less' else
            find_root(lambda p, k, n: _SimpleBinomial(n, p).sf(k-1) - alpha,
                      init, args=args).x)
    phigh = (xp.ones_like(k) if alternative == 'greater' else
             find_root(lambda p, k, n: _SimpleBinomial(n, p).cdf(k) - alpha,
                       init, args=args).x)

    plow = xp.where(k == 0, 0.0, plow)
    phigh = xp.where(k == n, 1.0, phigh)
    return plow, phigh


def _binom_wilson_conf_int(k, n, confidence_level, alternative, correction, *, xp):
    # This function assumes that the arguments have already been validated.
    # In particular, `alternative` must be one of 'two-sided', 'less' or
    # 'greater'.
    p = k / n
    if alternative == 'two-sided':
        z = ndtri(0.5 + 0.5*confidence_level)
    else:
        z = ndtri(confidence_level)

    # For reference, the formulas implemented here are from
    # Newcombe (1998) (ref. [3] in the proportion_ci docstring).
    denom = 2*(n + z**2)
    center = (2*n*p + z**2)/denom
    q = 1 - p

    if correction:
        with np.errstate(divide='ignore', invalid='ignore'):
            dlo = (1 + z*xp.sqrt(z**2 - 2 - 1/n + 4*p*(n*q + 1)))/denom
            dhi = (1 + z*xp.sqrt(z**2 + 2 - 1/n + 4*p*(n*q - 1)))/denom
    else:
        delta = z / denom * xp.sqrt(4*n*p*q + z**2)
        dlo, dhi = delta, delta

    lo = xp.where((k == 0) | (alternative == 'less'), 0.0, center - dlo)
    hi = xp.where((k == n) | (alternative == 'greater'), 1.0, center + dhi)
    return lo, hi


@xp_capabilities(skip_backends=[('dask.array', "")], cpu_only=True,
                 reason="binomial distribution ufuncs only available for NumPy",
                 extra_note="`alternative='two-sided'` is incompatible with JAX arrays.")
def binomtest(k, n, p=0.5, alternative='two-sided'):
    """
    Perform a test that the probability of success is p.

    The binomial test [1]_ is a test of the null hypothesis that the
    probability of success in a Bernoulli experiment is `p`.

    Details of the test can be found in many texts on statistics, such
    as section 24.5 of [2]_.

    The documentation is written as though the function accepts and returns Python
    scalars, but the function is vectorized to work elementwise with NumPy arrays.

    Parameters
    ----------
    k : int
        The number of successes.
    n : int
        The number of trials.
    p : float, optional
        The hypothesized probability of success, i.e. the expected
        proportion of successes.  The value must be in the interval
        ``0 <= p <= 1``. The default value is ``p = 0.5``.
    alternative : {'two-sided', 'greater', 'less'}, optional
        Indicates the alternative hypothesis. The default value is
        'two-sided'.

    Returns
    -------
    result : `~scipy.stats._result_classes.BinomTestResult` instance
        The return value is an object with the following attributes:

        k : int
            The number of successes (copied from `binomtest` input).
        n : int
            The number of trials (copied from `binomtest` input).
        alternative : str
            Indicates the alternative hypothesis specified in the input
            to `binomtest`.  It will be one of ``'two-sided'``, ``'greater'``,
            or ``'less'``.
        statistic : float
            The estimate of the proportion of successes.
        pvalue : float
            The p-value of the hypothesis test.

        The object has the following methods:

        proportion_ci(confidence_level=0.95, method='exact') :
            Compute the confidence interval for ``statistic``.

    Notes
    -----
    .. versionadded:: 1.7.0

    References
    ----------
    .. [1] Binomial test, https://en.wikipedia.org/wiki/Binomial_test
    .. [2] Jerrold H. Zar, Biostatistical Analysis (fifth edition),
           Prentice Hall, Upper Saddle River, New Jersey USA (2010)

    Examples
    --------
    >>> from scipy.stats import binomtest

    A car manufacturer claims that no more than 10% of their cars are unsafe.
    15 cars are inspected for safety, 3 were found to be unsafe. Test the
    manufacturer's claim:

    >>> result = binomtest(3, n=15, p=0.1, alternative='greater')
    >>> result.pvalue
    0.18406106910639114

    The null hypothesis cannot be rejected at the 5% level of significance
    because the returned p-value is greater than the critical value of 5%.

    The test statistic is equal to the estimated proportion, which is simply
    ``3/15``:

    >>> result.statistic
    0.2

    We can use the `proportion_ci()` method of the result to compute the
    confidence interval of the estimate:

    >>> result.proportion_ci(confidence_level=0.95)
    ConfidenceInterval(low=0.05684686759024681, high=1.0)

    """
    xp = array_namespace(k, n, p)
    k, n, p = xp_promote(k, n, p, force_floating=True, broadcast=True, xp=xp)
    k_valid = (k >= 0) & (k <= n) & (k == xp.floor(k))
    n_valid = (n >= 1) & (n == xp.floor(n))
    p_valid = (p >= 0) & (p <= 1)
    valid = k_valid & n_valid & p_valid
    k = xp.where(valid, k, xp.nan)
    n = xp.where(valid, n, xp.nan)
    p = xp.where(valid, p, xp.nan)

    if alternative not in ('two-sided', 'less', 'greater'):
        raise ValueError(f"alternative ('{alternative}') not recognized; \n"
                         "must be 'two-sided', 'less' or 'greater'")

    B = _SimpleBinomial(n, p)
    if alternative == 'less':
        pval = B.cdf(k)
    elif alternative == 'greater':
        pval = B.sf(k - 1)
    else:
        if is_jax(xp):
            message = "`alternative='two-sided'` is incompatible with JAX arrays."
            raise ValueError(message)

        # alternative is 'two-sided'
        d = B.pmf(k)
        rerr = 1 + 1e-7

        def k_lt_pn(d, k, p, n):
            B = _SimpleBinomial(n, p)
            ix = _binary_search_for_binom_tst(lambda x1: -B.pmf(x1), -d*rerr,
                                              xp.ceil(p * n), n, xp=xp)
            # y is the number of terms between mode and n that are <= d*rerr.
            # ix gave us the first term where a(ix) <= d*rerr < a(ix-1)
            # if the first equality doesn't hold, y=n-ix. Otherwise, we
            # need to include ix as well as the equality holds. Note that
            # the equality will hold in very very rare situations due to rerr.
            y = n - ix + xp.asarray(d*rerr == B.pmf(ix), dtype=ix.dtype)
            pval = B.cdf(k) + B.sf(n - y)
            return pval

        def k_gte_pn(d, k, p, n):
            B = _SimpleBinomial(n, p)
            ix = _binary_search_for_binom_tst(B.pmf, d*rerr,
                                              xp.zeros_like(n), xp.floor(p * n), xp=xp)
            # y is the number of terms between 0 and mode that are <= d*rerr.
            # we need to add a 1 to account for the 0 index.
            # For comparing this with old behavior, see
            # tst_binary_srch_for_binom_tst method in test_morestats.
            y = ix + 1
            pval = B.cdf(y-1) + B.sf(k-1)
            return pval

        pval = xpx.apply_where(k < p*n, (d, k, p, n), k_lt_pn,  k_gte_pn)
        # xp.minimum(1.0, pval) but for data-apis/array-api-compat#271
        pval = xp.minimum(xp.asarray(1.0, dtype=pval.dtype), pval)

    statistic = xp.where(valid, k/n, xp.nan)
    pval = xp.where(valid, pval, xp.nan)
    if statistic.ndim == 0:
        k, n, statistic, pval = k[()], n[()], statistic[()], pval[()]

    result = BinomTestResult(k=k, n=n, alternative=alternative,
                             statistic=statistic, pvalue=pval, xp=xp)
    return result


def _binary_search_for_binom_tst(a, d, lo, hi, *, xp):
    """
    Conducts an implicit binary search on a function specified by `a`.

    Meant to be used on the binomial PMF for the case of two-sided tests
    to obtain the value on the other side of the mode where the tail
    probability should be computed. The values on either side of
    the mode are always in order, meaning binary search is applicable.

    Parameters
    ----------
    a : callable
      The function over which to perform binary search. Its values
      for inputs lo and hi should be in ascending order.
    d : float
      The value to search.
    lo : int
      The lower end of range to search.
    hi : int
      The higher end of the range to search.

    Returns
    -------
    int
      The index, i between lo and hi
      such that a(i)<=d<a(i+1)
    """
    d = xp.asarray(d, copy=True)
    lo = xp.asarray(lo, copy=True)
    hi = xp.asarray(hi, copy=True)
    while xp.any(lo < hi):
        mid = lo + (hi-lo)//2
        midval = a(mid)

        i_lt = midval < d
        lo = xpx.at(lo)[i_lt].set(mid[i_lt] + 1)

        i_gt = midval > d
        hi = xpx.at(hi)[i_gt].set(mid[i_gt] - 1)

        i_eq = (midval == d)
        mid_i_eq = mid[i_eq]
        lo = xpx.at(lo)[i_eq].set(mid_i_eq)
        hi = xpx.at(hi)[i_eq].set(mid_i_eq)

    return xp.where(a(lo) <= d, lo, lo-1)


class _SimpleBinomial:
    # A very simple, array-API compatible binomial distribution for use in
    # hypothesis tests. May be replaced by new infrastructure Binomial
    # distribution in due time.
    def __init__(self, n, p, *, xp=None):
        xp = array_namespace(n, p) if xp is None else xp
        self.n = n
        self.p = p
        self.xp = xp

    def f(self, x, fun):
        xp = self.xp
        return xpx.lazy_apply(fun, x, self.n, self.p, as_numpy=True, xp=xp)

    def cdf(self, x):
        return self.xp.where(x >= 0, self.f(self.xp.floor(x), scu._binom_cdf), 0.0)

    def sf(self, x):
        return self.xp.where(x >= 0, self.f(self.xp.floor(x), scu._binom_sf), 1.0)

    def ppf(self, x):
        return self.f(x, scu._binom_ppf)

    def isf(self, x):
        return self.f(x, scu._binom_isf)

    def pmf(self, x):
        return self.f(self.xp.floor(x), scu._binom_pmf)



import itertools as it
import math
import pytest

import numpy as np

from scipy._lib._array_api import (is_array_api_strict, make_xp_test_case,
                                   xp_default_dtype, xp_device)
from scipy._lib._array_api_no_0d import (xp_assert_equal, xp_assert_close,
                                         xp_assert_less)

from scipy.special import log_softmax, logsumexp, softmax
from scipy.special._logsumexp import _wrap_radians


dtypes = ['float32', 'float64', 'int32', 'int64', 'complex64', 'complex128']
integral_dtypes = ['int32', 'int64']


@pytest.mark.uses_xp_capabilities(False, reason="private")
def test_wrap_radians(xp):
    x = xp.asarray([-math.pi-1, -math.pi, -1, -1e-300,
                    0, 1e-300, 1, math.pi, math.pi+1])
    ref = xp.asarray([math.pi-1, math.pi, -1, -1e-300,
                    0, 1e-300, 1, math.pi, -math.pi+1])
    res = _wrap_radians(x, xp=xp)
    xp_assert_close(res, ref, atol=0)


# numpy warning filters don't work for dask (dask/dask#3245)
# (also we should not expect the numpy warning filter to work for any Array API
# library)
@pytest.mark.filterwarnings("ignore:invalid value encountered:RuntimeWarning")
@pytest.mark.filterwarnings("ignore:divide by zero encountered:RuntimeWarning")
@pytest.mark.filterwarnings("ignore:overflow encountered:RuntimeWarning")
@make_xp_test_case(logsumexp)
class TestLogSumExp:
    def test_logsumexp(self, xp):
        # Test with zero-size array
        a = xp.asarray([])
        desired = xp.asarray(-xp.inf)
        xp_assert_equal(logsumexp(a), desired)

        # Test whether logsumexp() function correctly handles large inputs.
        a = xp.arange(200., dtype=xp.float64)
        desired = xp.log(xp.sum(xp.exp(a)))
        xp_assert_close(logsumexp(a), desired)

        # Now test with large numbers
        b = xp.asarray([1000., 1000.])
        desired = xp.asarray(1000.0 + math.log(2.0))
        xp_assert_close(logsumexp(b), desired)

        n = 1000
        b = xp.full((n,), 10000)
        desired = xp.asarray(10000.0 + math.log(n))
        xp_assert_close(logsumexp(b), desired)

        x = xp.asarray([1e-40] * 1000000)
        logx = xp.log(x)
        X = xp.stack([x, x])
        logX = xp.stack([logx, logx])
        xp_assert_close(xp.exp(logsumexp(logX)), xp.sum(X))
        xp_assert_close(xp.exp(logsumexp(logX, axis=0)), xp.sum(X, axis=0))
        xp_assert_close(xp.exp(logsumexp(logX, axis=1)), xp.sum(X, axis=1))

        # Handling special values properly
        inf = xp.asarray([xp.inf])
        nan = xp.asarray([xp.nan])
        xp_assert_equal(logsumexp(inf), inf[0])
        xp_assert_equal(logsumexp(-inf), -inf[0])
        xp_assert_equal(logsumexp(nan), nan[0])
        xp_assert_equal(logsumexp(xp.asarray([-xp.inf, -xp.inf])), -inf[0])

        # Handling an array with different magnitudes on the axes
        a = xp.asarray([[1e10, 1e-10],
                        [-1e10, -np.inf]])
        ref = xp.asarray([1e10, -1e10])
        xp_assert_close(logsumexp(a, axis=-1), ref)

        # Test keeping dimensions
        ref = xp.expand_dims(ref, axis=-1)
        xp_assert_close(logsumexp(a, axis=-1, keepdims=True), ref)

        # Test multiple axes
        xp_assert_close(logsumexp(a, axis=(-1, -2)), xp.asarray(1e10))

    def test_logsumexp_b(self, xp):
        a = xp.arange(200., dtype=xp.float64)
        b = xp.arange(200., 0., -1.)
        desired = xp.log(xp.sum(b*xp.exp(a)))
        xp_assert_close(logsumexp(a, b=b), desired)

        a = xp.asarray([1000, 1000])
        b = xp.asarray([1.2, 1.2])
        desired = xp.asarray(1000 + math.log(2 * 1.2))
        xp_assert_close(logsumexp(a, b=b), desired)

        x = xp.asarray([1e-40] * 100000)
        b = xp.linspace(1, 1000, 100000)
        logx = xp.log(x)
        X = xp.stack((x, x))
        logX = xp.stack((logx, logx))
        B = xp.stack((b, b))
        xp_assert_close(xp.exp(logsumexp(logX, b=B)), xp.sum(B * X))
        xp_assert_close(xp.exp(logsumexp(logX, b=B, axis=0)), xp.sum(B * X, axis=0))
        xp_assert_close(xp.exp(logsumexp(logX, b=B, axis=1)), xp.sum(B * X, axis=1))

    def test_logsumexp_sign(self, xp):
        a = xp.asarray([1, 1, 1])
        b = xp.asarray([1, -1, -1])

        r, s = logsumexp(a, b=b, return_sign=True)
        xp_assert_close(r, xp.asarray(1.))
        xp_assert_equal(s, xp.asarray(-1.))

    def test_logsumexp_sign_zero(self, xp):
        a = xp.asarray([1, 1])
        b = xp.asarray([1, -1])

        r, s = logsumexp(a, b=b, return_sign=True)
        assert not xp.isfinite(r)
        assert not xp.isnan(r)
        assert r < 0
        assert s == 0

    def test_logsumexp_sign_shape(self, xp):
        a = xp.ones((1, 2, 3, 4))
        b = xp.ones_like(a)

        r, s = logsumexp(a, axis=2, b=b, return_sign=True)
        assert r.shape == s.shape == (1, 2, 4)

        r, s = logsumexp(a, axis=(1, 3), b=b, return_sign=True)
        assert r.shape == s.shape == (1,3)

    def test_logsumexp_complex_sign(self, xp):
        a = xp.asarray([1 + 1j, 2 - 1j, -2 + 3j])

        r, s = logsumexp(a, return_sign=True)

        expected_sumexp = xp.sum(xp.exp(a))
        # This is the numpy>=2.0 convention for np.sign
        expected_sign = expected_sumexp / xp.abs(expected_sumexp)

        xp_assert_close(s, expected_sign)
        xp_assert_close(s * xp.exp(r), expected_sumexp)

    def test_logsumexp_shape(self, xp):
        a = xp.ones((1, 2, 3, 4))
        b = xp.ones_like(a)

        r = logsumexp(a, axis=2, b=b)
        assert r.shape == (1, 2, 4)

        r = logsumexp(a, axis=(1, 3), b=b)
        assert r.shape == (1, 3)

    def test_logsumexp_b_zero(self, xp):
        a = xp.asarray([1, 10000])
        b = xp.asarray([1, 0])

        xp_assert_close(logsumexp(a, b=b), xp.asarray(1.))

    def test_logsumexp_b_shape(self, xp):
        a = xp.zeros((4, 1, 2, 1))
        b = xp.ones((3, 1, 5))

        logsumexp(a, b=b)

    @pytest.mark.parametrize('arg', (1, [1, 2, 3]))
    def test_xp_invalid_input(self, arg):
        assert logsumexp(arg) == logsumexp(np.asarray(np.atleast_1d(arg)))

    def test_array_like(self):
        a = [1000, 1000]
        desired = np.asarray(1000.0 + math.log(2.0))
        xp_assert_close(logsumexp(a), desired)

    @pytest.mark.parametrize('dtype', dtypes)
    def test_dtypes_a(self, dtype, xp):
        dtype = getattr(xp, dtype)
        a = xp.asarray([1000., 1000.], dtype=dtype)
        desired_dtype = (xp.asarray(1.).dtype if xp.isdtype(dtype, 'integral')
                         else dtype)  # true for all libraries tested
        desired = xp.asarray(1000.0 + math.log(2.0), dtype=desired_dtype)
        xp_assert_close(logsumexp(a), desired)

    @pytest.mark.parametrize('dtype_a', dtypes)
    @pytest.mark.parametrize('dtype_b', dtypes)
    def test_dtypes_ab(self, dtype_a, dtype_b, xp):
        xp_dtype_a = getattr(xp, dtype_a)
        xp_dtype_b = getattr(xp, dtype_b)
        a = xp.asarray([2, 1], dtype=xp_dtype_a)
        b = xp.asarray([1, -1], dtype=xp_dtype_b)
        if is_array_api_strict(xp):
            # special-case for `TypeError: array_api_strict.float32 and
            # and array_api_strict.int64 cannot be type promoted together`
            xp_float_dtypes = [dtype for dtype in [xp_dtype_a, xp_dtype_b]
                               if not xp.isdtype(dtype, 'integral')]
            if len(xp_float_dtypes) < 2:  # at least one is integral
                xp_float_dtypes.append(xp.asarray(1.).dtype)
            desired_dtype = xp.result_type(*xp_float_dtypes)
        else:
            desired_dtype = xp.result_type(xp_dtype_a, xp_dtype_b)
            if xp.isdtype(desired_dtype, 'integral'):
               desired_dtype = xp_default_dtype(xp)
        desired = xp.asarray(math.log(math.exp(2) - math.exp(1)), dtype=desired_dtype)
        xp_assert_close(logsumexp(a, b=b), desired)

    def test_gh18295(self, xp):
        # gh-18295 noted loss of precision when real part of one element is much
        # larger than the rest. Check that this is resolved.
        a = xp.asarray([0.0, -40.0])
        res = logsumexp(a)
        ref = xp.logaddexp(a[0], a[1])
        xp_assert_close(res, ref)

    @pytest.mark.parametrize('dtype', ['complex64', 'complex128'])
    def test_gh21610(self, xp, dtype):
        # gh-21610 noted that `logsumexp` could return imaginary components
        # outside the range (-pi, pi]. Check that this is resolved.
        # While working on this, I noticed that all other tests passed even
        # when the imaginary component of the result was zero. This suggested
        # the need of a stronger test with imaginary dtype.
        rng = np.random.default_rng(324984329582349862)
        dtype = getattr(xp, dtype)
        shape = (10, 100)
        x = rng.uniform(1, 40, shape) + 1.j * rng.uniform(1, 40, shape)
        x = xp.asarray(x, dtype=dtype)

        res = logsumexp(x, axis=1)
        ref = xp.log(xp.sum(xp.exp(x), axis=1))
        max = xp.full_like(xp.imag(res), xp.pi)
        xp_assert_less(xp.abs(xp.imag(res)), max)
        xp_assert_close(res, ref)

        out, sgn = logsumexp(x, return_sign=True, axis=1)
        ref = xp.sum(xp.exp(x), axis=1)
        xp_assert_less(xp.abs(xp.imag(sgn)), max)
        xp_assert_close(out, xp.real(xp.log(ref)))
        xp_assert_close(sgn, ref/xp.abs(ref))

    def test_gh21709_small_imaginary(self, xp):
        # Test that `logsumexp` does not lose relative precision of
        # small imaginary components
        x = xp.asarray([0, 0.+2.2204460492503132e-17j])
        res = logsumexp(x)
        # from mpmath import mp
        # mp.dps = 100
        # x, y = mp.mpc(0), mp.mpc('0', '2.2204460492503132e-17')
        # ref = complex(mp.log(mp.exp(x) + mp.exp(y)))
        ref = xp.asarray(0.6931471805599453+1.1102230246251566e-17j)
        xp_assert_close(xp.real(res), xp.real(ref))
        xp_assert_close(xp.imag(res), xp.imag(ref), atol=0, rtol=1e-15)


    @pytest.mark.parametrize('x,y', list(it.product(
        [
            -np.inf,
            np.inf,
            complex(-np.inf, 0.),
            complex(-np.inf, -0.),
            complex(-np.inf, np.inf),
            complex(-np.inf, -np.inf),
            complex(np.inf, 0.),
            complex(np.inf, -0.),
            complex(np.inf, np.inf),
            complex(np.inf, -np.inf),
            # Phase in each quadrant.
            complex(-np.inf, 0.7533),
            complex(-np.inf, 2.3562),
            complex(-np.inf, 3.9270),
            complex(-np.inf, 5.4978),
            complex(np.inf, 0.7533),
            complex(np.inf, 2.3562),
            complex(np.inf, 3.9270),
            complex(np.inf, 5.4978),
        ], repeat=2)
    ))
    def test_gh22601_infinite_elements(self, x, y, xp):
        # Test that `logsumexp` does reasonable things in the presence of
        # real and complex infinities.
        res = logsumexp(xp.asarray([x, y]))
        ref = xp.log(xp.sum(xp.exp(xp.asarray([x, y]))))
        xp_assert_equal(res, ref)

    def test_no_writeback(self, xp):
        """Test that logsumexp doesn't accidentally write back to its parameters."""
        a = xp.asarray([5., 4.])
        b = xp.asarray([3., 2.])
        logsumexp(a)
        logsumexp(a, b=b)
        xp_assert_equal(a, xp.asarray([5., 4.]))
        xp_assert_equal(b, xp.asarray([3., 2.]))

    @pytest.mark.parametrize("x_raw", [1.0, 1.0j, []])
    def test_device(self, x_raw, xp, devices):
        """Test input device propagation to output."""
        for d in devices:
            x = xp.asarray(x_raw, device=d)
            assert xp_device(logsumexp(x)) == xp_device(x)
            assert xp_device(logsumexp(x, b=x)) == xp_device(x)

    def test_gh22903(self, xp):
        # gh-22903 reported that `logsumexp` produced NaN where the weight associated
        # with the max magnitude element was negative and `return_sign=False`, even if
        # the net result should be the log of a positive number.

        # result is log of positive number
        a = xp.asarray([3.06409428, 0.37251854, 3.87471931])
        b = xp.asarray([1.88190708, 2.84174795, -0.85016884])
        xp_assert_close(logsumexp(a, b=b), logsumexp(a, b=b, return_sign=True)[0])

        # result is log of negative number
        b = xp.asarray([1.88190708, 2.84174795, -3.85016884])
        xp_assert_close(logsumexp(a, b=b), xp.asarray(xp.nan))

    @pytest.mark.parametrize("a, b, sign_ref",
                             [([np.inf], None, 1.),
                              ([np.inf], [-1.], -1.)])
    def test_gh23548(self, xp, a, b, sign_ref):
        # gh-23548 reported that `logsumexp` with `return_sign=True` returned a sign
        # of NaN with infinite reals
        a, b = xp.asarray(a), xp.asarray(b) if b is not None else None
        val, sign = logsumexp(a, b=b, return_sign=True)
        assert xp.isinf(val)
        xp_assert_equal(sign, xp.asarray(sign_ref))


@make_xp_test_case(softmax)
class TestSoftmax:
    def test_softmax_fixtures(self, xp):
        xp_assert_close(softmax(xp.asarray([1000., 0., 0., 0.])),
                        xp.asarray([1., 0., 0., 0.]), rtol=1e-13)
        xp_assert_close(softmax(xp.asarray([1., 1.])),
                        xp.asarray([.5, .5]), rtol=1e-13)
        xp_assert_close(softmax(xp.asarray([0., 1.])),
                        xp.asarray([1., np.e])/(1 + np.e),
                        rtol=1e-13)

        # Expected value computed using mpmath (with mpmath.mp.dps = 200) and then
        # converted to float.
        x = xp.arange(4, dtype=xp.float64)
        expected = xp.asarray([0.03205860328008499,
                               0.08714431874203256,
                               0.23688281808991013,
                               0.6439142598879722], dtype=xp.float64)

        xp_assert_close(softmax(x), expected, rtol=1e-13)

        # Translation property.  If all the values are changed by the same amount,
        # the softmax result does not change.
        xp_assert_close(softmax(x + 100), expected, rtol=1e-13)

        # When axis=None, softmax operates on the entire array, and preserves
        # the shape.
        xp_assert_close(softmax(xp.reshape(x, (2, 2))),
                        xp.reshape(expected, (2, 2)), rtol=1e-13)

    def test_softmax_multi_axes(self, xp):
        xp_assert_close(softmax(xp.asarray([[1000., 0.], [1000., 0.]]), axis=0),
                        xp.asarray([[.5, .5], [.5, .5]]), rtol=1e-13)
        xp_assert_close(softmax(xp.asarray([[1000., 0.], [1000., 0.]]), axis=1),
                        xp.asarray([[1., 0.], [1., 0.]]), rtol=1e-13)

        # Expected value computed using mpmath (with mpmath.mp.dps = 200) and then
        # converted to float.
        x = xp.asarray([[-25.,   0.,  25.,  50.],
                        [  1., 325., 749., 750.]])
        expected = xp.asarray([[2.678636961770877e-33,
                                1.9287498479371314e-22,
                                1.3887943864771144e-11,
                                0.999999999986112],
                               [0.0,
                                1.9444526359919372e-185,
                                0.2689414213699951,
                                0.7310585786300048]])
        xp_assert_close(softmax(x, axis=1), expected, rtol=1e-13)
        xp_assert_close(softmax(x.T, axis=0), expected.T, rtol=1e-13)

        # 3-d input, with a tuple for the axis.
        x3d = xp.reshape(x, (2, 2, 2))
        xp_assert_close(softmax(x3d, axis=(1, 2)),
                        xp.reshape(expected, (2, 2, 2)), rtol=1e-13)

    @pytest.mark.xfail_xp_backends("array_api_strict", reason="int->float promotion")
    def test_softmax_int_array(self, xp):
        xp_assert_close(softmax(xp.asarray([1000, 0, 0, 0])),
                        xp.asarray([1., 0., 0., 0.]), rtol=1e-13)

    def test_softmax_scalar(self):
        xp_assert_close(softmax(1000), np.asarray(1.), rtol=1e-13)

    def test_softmax_array_like(self):
        xp_assert_close(softmax([1000, 0, 0, 0]),
                        np.asarray([1., 0., 0., 0.]), rtol=1e-13)


@make_xp_test_case(log_softmax)
class TestLogSoftmax:
    def test_log_softmax_basic(self, xp):
        xp_assert_close(log_softmax(xp.asarray([1000., 1.])),
                        xp.asarray([0., -999.]), rtol=1e-13)

    @pytest.mark.xfail_xp_backends("array_api_strict", reason="int->float promotion")
    def test_log_softmax_int_array(self, xp):
        xp_assert_close(log_softmax(xp.asarray([1000, 1])),
                        xp.asarray([0., -999.]), rtol=1e-13)

    def test_log_softmax_scalar(self):
        xp_assert_close(log_softmax(1.0), 0.0, rtol=1e-13)

    def test_log_softmax_array_like(self):
        xp_assert_close(log_softmax([1000, 1]),
                        np.asarray([0., -999.]), rtol=1e-13)

    @staticmethod
    def data_1d(xp):
        x = xp.arange(4, dtype=xp.float64)
        # Expected value computed using mpmath (with mpmath.mp.dps = 200)
        expect = [-3.4401896985611953,
                  -2.4401896985611953,
                  -1.4401896985611953,
                  -0.44018969856119533]
        return x, xp.asarray(expect, dtype=xp.float64)

    @staticmethod
    def data_2d(xp):
        x = xp.reshape(xp.arange(8, dtype=xp.float64), (2, 4))

        # Expected value computed using mpmath (with mpmath.mp.dps = 200)
        expect = [[-3.4401896985611953,
                   -2.4401896985611953,
                   -1.4401896985611953,
                   -0.44018969856119533],
                  [-3.4401896985611953,
                   -2.4401896985611953,
                   -1.4401896985611953,
                   -0.44018969856119533]]
        return x, xp.asarray(expect, dtype=xp.float64)

    @pytest.mark.parametrize("offset", [0, 100])
    def test_log_softmax_translation(self, offset, xp):
        # Translation property.  If all the values are changed by the same amount,
        # the softmax result does not change.
        x, expect = self.data_1d(xp)
        x += offset
        xp_assert_close(log_softmax(x), expect, rtol=1e-13)

    def test_log_softmax_noneaxis(self, xp):
        # When axis=None, softmax operates on the entire array, and preserves
        # the shape.
        x, expect = self.data_1d(xp)
        x = xp.reshape(x, (2, 2))
        expect = xp.reshape(expect, (2, 2))
        xp_assert_close(log_softmax(x), expect, rtol=1e-13)

    @pytest.mark.parametrize('axis_2d, expected_2d', [
        (0, np.log(0.5) * np.ones((2, 2))),
        (1, [[0., -999.], [0., -999.]]),
    ])
    def test_axes(self, axis_2d, expected_2d, xp):
        x = xp.asarray([[1000., 1.], [1000., 1.]])
        xp_assert_close(log_softmax(x, axis=axis_2d),
                        xp.asarray(expected_2d, dtype=x.dtype), rtol=1e-13)

    def test_log_softmax_2d_axis1(self, xp):
        x, expect = self.data_2d(xp)
        xp_assert_close(log_softmax(x, axis=1), expect, rtol=1e-13)

    def test_log_softmax_2d_axis0(self, xp):
        x, expect = self.data_2d(xp)
        xp_assert_close(log_softmax(x.T, axis=0), expect.T, rtol=1e-13)

    def test_log_softmax_3d(self, xp):
        # 3D input, with a tuple for the axis.
        x, expect = self.data_2d(xp)
        x = xp.reshape(x, (2, 2, 2))
        expect = xp.reshape(expect, (2, 2, 2))
        xp_assert_close(log_softmax(x, axis=(1, 2)), expect, rtol=1e-13)
